#
#	THIS FILE IS PART OF THE JOKOSHER PROJECT AND LICENSED UNDER THE GPL. SEE
#	THE 'COPYING' FILE FOR DETAILS
#
#	This module shows the Extensions Manager dialog which is used to add, remove
#	and configure Jokosher's Extensions.
#
#-------------------------------------------------------------------------------

import gtk.glade, pango
import Globals, Extension
import gettext
import os
import gettext
_ = gettext.gettext

class ExtensionManagerDialog:
	"""
	This module shows the Extensions Manager dialog which is used to add, remove
	and configure Jokosher's Extensions.
	"""
	def __init__(self, parent):
		"""
		Creates a new instance of ExtensionManagerDialog.
		
		Parameters:
			parent -- the parent MainApp Jokosher window.
		"""
		self.parent = parent
		
		self.wTree = gtk.glade.XML(Globals.GLADE_PATH, "ExtensionManagerDialog")
		
		signals = {
			"on_Close_clicked" : self.OnClose,
			"on_Add_clicked" : self.OnAdd,
			"on_Remove_clicked": self.OnRemove,
			"on_Treeview_selected": self.OnSelect,
			"on_Preferences_clicked": self.OnPreferences
		}
		self.wTree.signal_autoconnect(signals)
		
		self.dlg = self.wTree.get_widget("ExtensionManagerDialog")
		self.tree = self.wTree.get_widget("treeviewExtensions")
		self.eb = self.wTree.get_widget("descriptionEventBox")
		self.description = self.wTree.get_widget("ExtensionDescription")
		self.prefs_button = self.wTree.get_widget("buttonPreferences")
		
		self.AddColumn("Enabled", 0, 'toggle')
		self.AddColumn("Name", 1, 'text', 30)
		self.AddColumn("Version", 3, 'text', 7)

		self.model = gtk.ListStore(bool, str, str, str, str, bool)
		self.tree.set_model(self.model)

		for extension in self.parent.extensionManager.GetExtensions():
			self.model.append((extension["enabled"], extension["name"], extension["description"], extension["version"], extension["filename"], extension["preferences"]))


		self.dlg.set_transient_for(self.parent.window)
		self.dlg.show()
	
	#_____________________________________________________________________

	def OnClose(self, button):
		"""
		Destroys the dialog when the close button is pressed.
		It also saves the Extensions's settings.
		
		Parameters:
			button -- reserved for GTK callbacks, don't use it explicitly.
		"""
		self.dlg.destroy()
		Globals.settings.write()
	#_____________________________________________________________________

	def AddColumn(self, title, modelId, cell_renderer='text', cell_width=20):
		"""
		Adds a display column to the ExtensionManager dialog.
		
		Parameters:
			title -- title of the column.
			modelId -- index of the column within the TreeView model.
			cell_renderer -- type of render to use:
							'text' = renders text into the TreeView.
							'toggle' = renders toggle buttons into the TreeView.
			cell_width -- width in pixels of the column.
		"""
		if cell_renderer == 'toggle':
			renderer = gtk.CellRendererToggle()
			renderer.set_property('activatable', True)
			column = gtk.TreeViewColumn(title, renderer, active=modelId)
			renderer.connect('toggled', self.ToggleEnabled)
		
		else:
			renderer = gtk.CellRendererText()
			renderer.set_property('ellipsize', pango.ELLIPSIZE_END)
			renderer.set_property('ellipsize-set', True)
			renderer.set_property('width-chars', cell_width)
			column = gtk.TreeViewColumn(title, renderer, text=modelId)
			column.set_property('resizable', True)

		self.tree.append_column(column)
	
	#_____________________________________________________________________

	def ToggleEnabled(self, cell, path):
		"""
		Enables/disables an Extension.
		
		Parameters:
			cell -- reserved for GTK callbacks, don't use it explicitly.
			path -- path to the Extension.
		"""
		self.model[path][0] = not self.model[path][0]
		
		iter = self.model.get_iter(path)
		name = self.model.get_value(iter, 1)
		filename = self.model.get_value(iter, 4)
		
		if self.model[path][0]:
			if name in Globals.settings.extensions['extensions_blacklist']:
				Globals.settings.extensions['extensions_blacklist'] = Globals.settings.extensions['extensions_blacklist'].replace(name, "")
				Globals.debug("extension-blacklist:", Globals.settings.extensions['extensions_blacklist'])
				for extension in self.parent.extensionManager.GetExtensions():
					if filename == extension['filename']:
						extension['enabled'] = True

				self.parent.extensionManager.StartExtension(filename)
		else:
			if name not in Globals.settings.extensions['extensions_blacklist']:
				Globals.settings.extensions['extensions_blacklist'] += name+" "
				self.parent.extensionManager.StopExtension(filename)
				
				for extension in self.parent.extensionManager.GetExtensions():
					if filename == extension['filename']:
						extension['enabled'] = False
	#_____________________________________________________________________

	def OnSelect(self, tree):
		"""
		When an Extension is selected, enables/disables the preferences button
		according to the reported Extension capabilities and updates the
		description label.
		
		Parameters:
			tree -- GTKTreeView holding the Extensions's representation.
		"""
		descr_text = ""
		selection = self.tree.get_selection().get_selected()[1]
		if selection:
			preferences = self.model.get_value(selection, 5)
			if preferences:
				self.prefs_button.set_sensitive(True)
			else:
				self.prefs_button.set_sensitive(False)
		
			descr_text = self.model.get_value(selection, 2)
		self.description.set_text(descr_text)
	#_____________________________________________________________________
	
	def OnAdd(self, button):
		"""
		Displays a dialog which allows the user to add an external Extension.
		If an error occurs, an error dialog is displayed.
		
		Parameters:
			button -- reserved for GTK callbacks, don't use it explicitly.
		"""
		chooser = gtk.FileChooserDialog((_('Choose a Jokosher Extension file')), None, 
				gtk.FILE_CHOOSER_ACTION_OPEN, (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL, gtk.STOCK_OPEN, gtk.RESPONSE_OK))
		chooser.set_default_response(gtk.RESPONSE_OK)
		chooser.set_transient_for(self.dlg)

		filter = gtk.FileFilter()
		filter.set_name(_("Jokosher Extension Files (*.py *.egg)"))
		filter.add_pattern("*.py")
		filter.add_pattern("*.egg")

		chooser.add_filter(filter)

		response = chooser.run()
			
		if response == gtk.RESPONSE_OK:
			install_dlg = gtk.MessageDialog(None, gtk.DIALOG_MODAL, gtk.MESSAGE_QUESTION, 
					gtk.BUTTONS_YES_NO, "Install extension to local jokosher extension directory?")
			install_response = install_dlg.run()
			
			if install_response == gtk.RESPONSE_YES:
				filename = chooser.get_filename()
				#TODO: redo the enable/disable stuff here
				
				if not self.parent.extensionManager.LoadExtensionFromFile(os.path.basename(filename), os.path.dirname(filename), local=True):
					dlg = gtk.MessageDialog(self.dlg,
						gtk.DIALOG_MODAL | gtk.DIALOG_DESTROY_WITH_PARENT,
						gtk.MESSAGE_ERROR,
						gtk.BUTTONS_CLOSE,
						_("There was a problem loading the extension"))
					dlg.run()
					dlg.destroy()
				self.UpdateModel()
				chooser.destroy()
			
			else:
				install_dlg.destroy()
			install_dlg.destroy()	
								
		chooser.destroy()
	
	#_____________________________________________________________________

	def OnRemove(self, button):
		"""
		Removes an Extension from the Extensions list.
		If an error occurs, an error dialog is displayed.
		
		Parameters:
			button -- reserved for GTK callbacks, don't use it explicitly.
		"""
		selection = self.tree.get_selection()
		row_selected = selection.get_selected()[1]
		filename = self.model.get_value(row_selected, 4)
		
		if row_selected:
			if Extension.EXTENSION_DIR_USER in filename:
				dlg = gtk.MessageDialog(None, gtk.DIALOG_MODAL, gtk.MESSAGE_QUESTION, 
						gtk.BUTTONS_YES_NO, "Remove the selected extension?")
				response = dlg.run()
				if response == gtk.RESPONSE_YES:
					dlg.destroy()
					if not self.parent.extensionManager.RemoveExtension(filename): #don't forget to change to RemoveExtenion() in ExtensionManager
						dlg2 = gtk.MessageDialog(self.dlg,
							gtk.DIALOG_MODAL | gtk.DIALOG_DESTROY_WITH_PARENT,
							gtk.MESSAGE_ERROR,
							gtk.BUTTONS_CLOSE,
							_("There was a problem removing the extension"))
						dlg2.run()
						dlg2.destroy()
						
					self.UpdateModel()
				
				dlg.destroy()

			else:
				dlg = gtk.MessageDialog(None, gtk.DIALOG_MODAL, gtk.MESSAGE_ERROR,
						gtk.BUTTONS_OK, "This extension was installed by your system. You cannot remove it.")
				dlg.run()
				dlg.destroy()
		else:
			dlg = gtk.MessageDialog(None, gtk.DIALOG_MODAL, gtk.MESSAGE_ERROR,
					gtk.BUTTONS_OK, "No extension selected!")
			dlg.run()
			dlg.destroy()

	#_____________________________________________________________________

	def OnPreferences(self, button):
		"""
		Displays the Extension's preferences dialog.
		If an error occurs, an error dialog is displayed.
		
		Parameters:
			button -- reserved for GTK callbacks, don't use it explicitly.
		"""
		selection = self.tree.get_selection().get_selected()[1]
		filename = self.model.get_value(selection, 4)

		if not self.parent.extensionManager.ExtensionPreferences(filename):
			dlg = gtk.MessageDialog(self.dlg,
							gtk.DIALOG_MODAL | gtk.DIALOG_DESTROY_WITH_PARENT,
							gtk.MESSAGE_ERROR,
							gtk.BUTTONS_CLOSE,
							_("An error occurred when trying to launch the preferences for the extension"))
			dlg.run()
			dlg.destroy()

	#_____________________________________________________________________

	def UpdateModel(self):
		"""
		Updates the ExtensionManagerDialog Extension list, to reflect changes
		in the currently available Extensions.
		"""
		selection = self.tree.get_selection()
		row_selected = selection.get_selected()[1]
		
		num_extensions = len(self.parent.extensionManager.loadedExtensions)
		num_model = len(self.model)
		
		if num_model < num_extensions:
			extension = self.parent.extensionManager.loadedExtensions[num_extensions-1]
			self.model.append((extension["enabled"], extension["name"], extension["description"], extension["version"], extension["filename"], extension["preferences"]))
		
		elif num_model > num_extensions:
			self.model.remove(row_selected)

	#_____________________________________________________________________



#
#	THIS FILE IS PART OF THE JOKOSHER PROJECT AND LICENSED UNDER THE GPL. SEE
#	THE 'COPYING' FILE FOR DETAILS
#
#	ProjectTemplate.py
#	
#	This module handles the saving and loading of project templates.
#	
#
#-------------------------------------------------------------------------------

import os
import gobject
import Globals
import xml.dom.minidom as xml
from Utils import LoadListFromXML, StoreListToXML

#=========================================================================

class ProjectTemplate(gobject.GObject):
	"""
	This class saves and loads templates and template information to disk.
	"""

	#the extension put on the end of template files
	TEMPLATE_EXT = "template"
	
	"""
	Signals:
		"template-update" -- This template has changed.
	"""
	
	__gsignals__ = {
		"template-update" : ( gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, () )
	}
	
	#_____________________________________________________________________

	def __init__(self):
		""" 
		Creates a new instance of ProjectTemplate.
		"""
		gobject.GObject.__init__(self)

	#_____________________________________________________________________

	def SaveTemplateFile(self, name, instrlist):
		""" 
		This method will write a template file to JOKOSHER_DATA_HOME/templates/.
		instrlist is a list containing type strings corresponding to each instrument.
		e.g. ["audiofile", "acousticguitar"].
	
		Parameters:
			name -- the name of the template file.
			instrlist -- a list containing the instrument type strings
		"""
		doc = xml.Document()
		head = doc.createElement("JokosherProjectTemplate")
		doc.appendChild(head)
		
		for typeString in instrlist:
			# create instrument tags for every instrument in the list
			instrtag = doc.createElement("Instrument")
			instrtag.setAttribute("type", typeString)
			head.appendChild(instrtag)
		
		if not name.endswith("." + self.TEMPLATE_EXT):
			name += "." + self.TEMPLATE_EXT
		namePath = os.path.join(Globals.TEMPLATES_PATH, name)
		try:
			try:
				filename = open(namePath, "w")
				filename.write(doc.toprettyxml())
			except IOError, e:
				Globals.debug("The template %s does not exist" % namePath)
		finally:
			filename.close()
			
		self.emit("template-update")

	#_____________________________________________________________________
	
	def DeleteTemplateFile(self, name):
		"""
		This method will delete a template file.
		
		Parameters:
			name -- the name of the template file which will be deleted.
		"""
		if not name.endswith("." + self.TEMPLATE_EXT):
			name += "." + self.TEMPLATE_EXT
		namePath = os.path.join(Globals.TEMPLATES_PATH, name)
		try:
			os.remove(namePath)
		except OSError, e:
			Globals.debug("Cannot remove template %s" % namePath)
			
		self.emit("template-update")

	#_____________________________________________________________________
	
	def __LoadInstrumentsFromTemplateFile(self, name):
		"""
		This method will return a list containing a list of instruments which are in the given template filename.
		
		Parameters:
			name -- the name of the template.
			
		Returns:
			instrlist -- a list containing a list of instruments, e.g. [["audiofile", "Audio File", "path_to_image"]].
		"""
		if not name.endswith("." + self.TEMPLATE_EXT):
			name += "." + self.TEMPLATE_EXT
		namePath = os.path.join(Globals.TEMPLATES_PATH, name)
		if os.path.exists(namePath):
			file = open(namePath, "r")
			doc = xml.parse(file)
			instrlist = []
			for node in doc.getElementsByTagName("Instrument"):
				typeString = str(node.getAttribute("type"))
				instrlist.append(typeString)
			return instrlist
		else:
			Globals.debug("The template %s does not exist" % namePath)
	
	#_____________________________________________________________________
	
	def __GetTemplateList(self):
		"""
		This method will return a list of template files in the template directory (JOKOSHER_DATA_HOME/templates/")
		
		Returns:
			filenames -- a list containing the names of file in the template directory.
		"""
		templist = []
		for files in os.listdir(Globals.TEMPLATES_PATH):
			filenames = files.split(".")[0]
			templist.append(filenames)
		return templist

	#_____________________________________________________________________

	def LoadDictionaryOfInstrumentsFromTemplateFile(self):
		"""
		This method will return a dictionary containing the the template file name (key) and a list
		of the associated instrument tupels.
		e.g. {"Rock" : [ ("electricguitar", "Electric Guitar", pixbufImage) ]}
		"""
		instrdict = {}
		cached = Globals.getCachedInstruments()
		for filename in self.__GetTemplateList():
			instrTuples = []
			typeStringList = self.__LoadInstrumentsFromTemplateFile(filename)
			for type in typeStringList:
				tuple_ = [x for x in cached if x[1] == type]
				if tuple_:
					name, type, pixbuf, path = tuple_[0]
					instrTuples.append( (name, type, pixbuf) )
					
			instrdict[filename] = instrTuples
		
		return instrdict
	
#=========================================================================	



#
#	THIS FILE IS PART OF THE JOKOSHER PROJECT AND LICENSED UNDER THE GPL. SEE
#	THE 'COPYING' FILE FOR DETAILS
#
#	TimeLineBar.py
#	
#	This module is the gtk.Frame which sits above the instruments and
#	holds the TImeLine as well as the click track button and the bpm label.
#
#-------------------------------------------------------------------------------

import gtk, gtk.glade
import gobject
import gettext
import os
import TimeLine
import Globals
import ui.EventLaneHSeparator as EventLaneHSeparator

# for the time signature combo
from math import log

_=gettext.gettext

class TimeLineBar(gtk.HBox):
	"""
	This class contains the TimeLine widget as well as the click track button and the bpm label in a gtk.Frame widget.
	"""
	#_____________________________________________________________________
	
	def __init__(self, project, mainview):
		"""
		Creates a new instance of TimeLineBar
		
		Parameters:
			project -- reference to Project (Project.py).
			mainview -- reference to MainApp (JokosherApp.py).
		"""
		gtk.HBox.__init__(self)
		
		self.project = project
		self.mainview = mainview
		self.timeline = TimeLine.TimeLine(self.project, mainview)
		self.Updating = False
		
		# add click / bpm / signature box
		self.clickbutton = gtk.VolumeButton()
		self.clickbutton.set_value(0)
		self.clickbutton.set_relief(gtk.RELIEF_NORMAL)
		self.clickbutton.set_property("size", gtk.ICON_SIZE_BUTTON)
		
		self.clickbutton_metronome_image = gtk.Image()
		self.clickbutton_metronome_image.set_from_file(os.path.join(Globals.IMAGE_PATH, "icon_click.png"))
		# use get_child() not get_image here because GtkScaleButton
		# uses gtk_container_add() in its internal implementation
		self.clickbutton_volume_image = self.clickbutton.get_child()
		self.clickbutton.set_image(self.clickbutton_metronome_image)
		
		image_size_group = gtk.SizeGroup(gtk.SIZE_GROUP_BOTH)
		image_size_group.add_widget(self.clickbutton_metronome_image)
		image_size_group.add_widget(self.clickbutton_volume_image)
		
		self.clicktip = gtk.Tooltips()
		self.clicktip.set_tip(self.clickbutton, _("Adjust volume of click track"), None)
					
		self.bpmeventbox = gtk.EventBox()
		self.bpmframe = gtk.Frame()
		self.bpmeventtip = gtk.Tooltips()
		self.bpmeventtip.set_tip(self.bpmeventbox, _("Beats per minute"), None)
		self.bpmframe.set_shadow_type(gtk.SHADOW_ETCHED_OUT)
		
		self.bpmlabel = gtk.Label()
		self.bpmlabel.set_use_markup(True)
		self.bpmlabel.set_markup("<b>%s</b>" % (self.project.bpm))
		self.bpmlabel.set_padding(5, 5)
		self.bpmeventbox.add(self.bpmlabel)
		self.bpmframe.add(self.bpmeventbox)
		
		self.bpmedit = gtk.SpinButton()
		self.bpmedit.set_range(1, 400)
		self.bpmedit.set_increments(1, 5)
		self.bpmedit.set_value(self.project.bpm)
		self.bpmedit.connect("activate", self.OnAcceptEditBPM)
		self.bpmedit.connect("key_press_event", self.OnEditBPMKey)
		self.bpmedit.connect_after("button-release-event", self.OnEditBPMClick)

		self.sigeventbox = gtk.EventBox()
		self.sigeventtip = gtk.Tooltips()
		self.sigeventtip.set_tip(self.sigeventbox, _("Time signature"), None)
		self.sigframe = gtk.Frame()
		self.sigframe.set_shadow_type(gtk.SHADOW_ETCHED_OUT)
		
		self.siglabel = gtk.Label()
		self.siglabel.set_use_markup(True)
		self.siglabel.set_markup("<b>%d/%d</b>" % (self.project.meter_nom, self.project.meter_denom))
		self.siglabel.set_padding(5, 5)
		self.sigeventbox.add(self.siglabel)
		self.sigframe.add(self.sigeventbox)
		self.sigeditPacked = False

		# set events
		self.clickbutton.connect("value-changed", self.OnClickTrackVolume)
		self.clickbutton.connect("enter_notify_event", self.OnClickButtonEnter)
		self.clickbutton.connect("leave_notify_event", self.OnClickButtonEnter)
		
		self.bpmeventbox.set_events(gtk.gdk.BUTTON_PRESS_MASK)
		self.bpmeventbox.connect("button_release_event", self.OnEditBPM)
		self.bpmeventbox.connect("enter_notify_event", self.OnMouseMoveBPM)
		self.bpmeventbox.connect("leave_notify_event", self.OnMouseMoveBPM)
		
		self.sigeventbox.set_events(gtk.gdk.BUTTON_PRESS_MASK)
		self.sigeventbox.connect("button_press_event", self.OnEditSig)
		self.sigeventbox.connect("enter_notify_event", self.OnMouseMoveTimeSig)
		self.sigeventbox.connect("leave_notify_event", self.OnMouseMoveTimeSig)
		
		self.project.connect("bpm", self.OnProjectBPMChange)
		self.project.connect("time-signature", self.OnProjectSigChange)
		self.project.connect("click-track", self.OnProjectClickTrackChange)

		self.sigDialog = None
		
		# ###########################################################
		
		self.headerhbox = gtk.HBox()
		self.headerhbox.set_border_width(2)
		self.headerhbox.set_spacing(5)
		self.headerhbox.pack_start(self.clickbutton, True, True)
		self.headerhbox.pack_start(self.bpmframe, True, True)
		self.headerhbox.pack_start(self.sigframe, True, True)
		
		self.headerVBox = gtk.VBox()
		self.headerVBox.pack_start(self.headerhbox, True, True)
		self.headerVBox.pack_start(gtk.HSeparator(), False, False)
		
		self.timelineVBox = gtk.VBox()
		self.timelineVBox.pack_start(self.timeline, True, True)
		separator = EventLaneHSeparator.EventLaneHSeparator(self.project, self.project.transport)
		self.timelineVBox.pack_start(separator, False, False)
		
		self.pack_start(self.headerVBox, False, False)
		self.pack_start(self.timelineVBox)	

	#_____________________________________________________________________
	
	def GetHeaderWidget(self):
		"""
			Returns the widget which is required to be aligned with the instrument headers.
		"""
		return self.headerVBox
	
	#_____________________________________________________________________
	
	def OnProjectBPMChange(self, project):
		"""
		Callback for when the BPM of the project changes. This method 
		will update the contents of the beats per minute box.
		
		Parameters:
			project -- The project that send the signal.
		"""
		#Do this outside the if statement so that it gets updated if someone else changes the bpm
		self.bpmlabel.set_use_markup(True)
		self.bpmlabel.set_markup("<b>%d</b>" % (self.project.bpm))
		
	#_____________________________________________________________________
	
	def OnProjectSigChange(self, project):
		"""
		Callback for when the BPM of the project changes. This method 
		will update the contents of the time signature label.
		
		Parameters:
			project -- The project that send the signal.
		"""
		self.siglabel.set_use_markup(True)
		self.siglabel.set_markup("<b>%d/%d</b>" % (self.project.meter_nom, self.project.meter_denom))
		
	#_____________________________________________________________________
	
	def OnProjectClickTrackChange(self, project, value):
		"""
		Callback for when the click track of the project it turned on
		or shut off. This method will update the button and tooltip state.
		
		Parameters:
			project -- The project that send the signal.
			value -- The new value of the click track volume.
		"""
		self.Updating = True
		self.clickbutton.set_value(value)
		self.Updating = False
		
	#_____________________________________________________________________
	
	def OnEditBPM(self, widget, event):
		"""
		Called when the user clicks the beats per minute box.
		This method will show a spin button widget with a value inside which the user can change.
		
		Parameters:
			widget -- reserved for GTK callbacks, don't use it explicitly.
			event -- reserved for GTK callbacks, don't use it explicitly.
		"""
		if event.type == gtk.gdk.BUTTON_RELEASE:
			startWidth = self.headerhbox.size_request()[0]
			
			if self.bpmeventbox.parent:
				self.bpmframe.remove(self.bpmeventbox)
				self.bpmframe.add(self.bpmedit)
				self.bpmedit.show()
				
				self.bpmedit.grab_add()
				self.bpmedit.grab_focus()

	#_____________________________________________________________________
	
	def OnAcceptEditBPM(self, widget=None, cancel=False):
		"""
		Called when the user finishes editing the beats per minute box.
		This method then updates the beats per minute value to the value the user 
		enters.
		
		Parameters:
			widget -- reserved for GTK callbacks, don't use it explicitly.
		"""
		self.bpmedit.grab_remove()
		
		if self.bpmedit.parent:
			self.bpmframe.remove(self.bpmedit)
			self.bpmframe.add(self.bpmeventbox)
			self.bpmframe.show_all()
			
			if not cancel:
				newbpm = self.bpmedit.get_value_as_int()
				self.project.SetBPM(float(newbpm))
		
	#_____________________________________________________________________
	
	def OnEditBPMClick(self, widget, event):
		"""
		Handles the button presses while doing a grab on the bpm edit spinbutton.
		If we get a mouse event, it means the user has clicked outside the spinbutton.
		So then we save what the user edited and go back to a label.
		"""
		if event.type == gtk.gdk.BUTTON_RELEASE:
			self.OnAcceptEditBPM()
			return True

	#_____________________________________________________________________

	def OnEditBPMKey(self, widget, event):
		"""
		Handles the key presses while editing the instrument name label.
		Used to make the escape key save the name and then return to normal mode.
		
		Parameters:
			widget -- GTK callback.
			event -- GTK callback.
		"""
		key = gtk.gdk.keyval_name(event.keyval)
		
		if key == "Escape":
			self.OnAcceptEditBPM(cancel=True)
			return True

	#_____________________________________________________________________

	def OnEditSig(self, widget, event):
		"""
		Called when the user clicks the time signature box.
		This method will show a dialog for changing the time signature.
		
		Parameters:
			widget -- reserved for GTK callbacks, don't use it explicitly.
			event -- reserved for GTK callbacks, don't use it explicitly.
		""" 
		if event.type == gtk.gdk.BUTTON_PRESS:
			if not self.sigDialog:
				self.sigDialogTree = gtk.glade.XML(Globals.GLADE_PATH, "TimeSignatureDialog")
		
				signals = {
					"on_spinBeats_value_changed" : self.OnNomValueChange,
					"on_comboValue_changed" : self.OnDenomValueChange,
					"on_buttonClose_clicked" : self.OnSigDialogClose
				}
				
				self.sigDialogTree.signal_autoconnect(signals)

				# references to GUI widgets
				self.sigDialog = self.sigDialogTree.get_widget("TimeSignatureDialog")
				self.spinBeats = self.sigDialogTree.get_widget("spinBeats")
				self.comboValue = self.sigDialogTree.get_widget("comboValue")
				
				# set some properties for the widgets
				self.sigDialog.set_icon(self.mainview.icon)
				self.spinBeats.set_value(self.project.meter_nom)
				self.comboValue.set_active(int(log(self.project.meter_denom, 2)))
			else:
				self.sigDialog.present()
			
	#_____________________________________________________________________
	
	def OnNomValueChange(self, spinButton):
		"""
		Updates the Project's beats per measure value.
		
		Parameters:
			combobox -- beats per measure combobox.
		"""
		self.project.SetMeter(int(spinButton.get_value()),
							  self.project.meter_denom)
		
	#_____________________________________________________________________
		
	def OnDenomValueChange(self, combobox):
		"""
		Updates the Project's beat value.
		
		Parameters:
			spinButton -- beat value spin button.
		"""
		self.project.SetMeter(self.project.meter_nom,
							  int(combobox.get_active_text()))
	
	#_____________________________________________________________________
	
	def OnSigDialogClose(self, button):
		"""
		Closes the time signature settings dialog.
		
		Parameters:
			button -- reserved for GTK callbacks, don't use it explicitly.
		"""
		self.sigDialog.destroy()
		self.sigDialog = None

	#_____________________________________________________________________
	
	def OnMouseMoveBPM(self, widget, event):
		"""
		Called when the mouse pointer enters or leaves the beats per minute box.
		This method changes the type of cursor if the mouse pointer is hovered over the beats per minute box.
				
		Parameters:
			widget -- reserved for GTK callbacks, don't use it explicitly.
			event -- reserved for GTK callbacks, don't use it explicitly.
		"""
		if not widget.window: 
			return
		
		if (event.type == gtk.gdk.ENTER_NOTIFY):
			widget.window.set_cursor(gtk.gdk.Cursor(gtk.gdk.XTERM))
		else:
			widget.window.set_cursor(None)
			
	#_____________________________________________________________________
	
	def OnMouseMoveTimeSig(self, widget, event):
		"""
		Called when the mouse pointer enters or leaves the beats per minute box.
		This method changes the type of cursor if the mouse pointer is hovered over the beats per minute box.
				
		Parameters:
			widget -- reserved for GTK callbacks, don't use it explicitly.
			event -- reserved for GTK callbacks, don't use it explicitly.
		"""
		if not widget.window: 
			return
		
		if (event.type == gtk.gdk.ENTER_NOTIFY):
			self.sigeventbox.set_state(gtk.STATE_PRELIGHT)
		else:
			self.sigeventbox.set_state(gtk.STATE_NORMAL)
			
	#_____________________________________________________________________

	def OnClickTrackVolume(self, widget, value):
		"""
		Called when the click button is clicked. This method will call the project
		to set the volume of the click track.
		""" 
		if not self.Updating:
			self.project.SetClickTrackVolume(value)
			
	#_____________________________________________________________________
	
	def OnClickButtonEnter(self, widget, event):
		if event.type == gtk.gdk.ENTER_NOTIFY:
			self.clickbutton.set_image(self.clickbutton_volume_image)
		elif event.type == gtk.gdk.LEAVE_NOTIFY:
			self.clickbutton.set_image(self.clickbutton_metronome_image)
		
	#_____________________________________________________________________
	
#=========================================================================



#
#	THIS FILE IS PART OF THE JOKOSHER PROJECT AND LICENSED UNDER THE GPL. SEE
#	THE 'COPYING' FILE FOR DETAILS
#
#	Contains a master VU widget and levels for all Instruments.
#
#-------------------------------------------------------------------------------
import pygtk
pygtk.require("2.0")
import gtk

from VUWidget import *
import gettext
_ = gettext.gettext

#=========================================================================

class MasterMixerStrip(gtk.Frame):
	"""
	Contains a master VU widget and levels for all Instruments.
	"""
	
	""" GTK widget name """
	__gtype_name__ = 'MasterMixerStrip'
	
	#_____________________________________________________________________
	
	def __init__(self, project, mixview, mainview):
		"""
		Contains a new instance of MasterMixerStrip.
		
		Parameters:
			project -- the currently active Project.
			mixview -- the mixing view object (CompactMixView).
			mainview -- the main Jokosher window (MainApp).
		"""
		gtk.Container.__init__(self)
		self.project = project
		self.mixview = mixview
		self.mainview = mainview
		self.Updating = False
		
		self.vbox = gtk.VBox()
		self.add(self.vbox)

		self.label = gtk.Label(_("Master Volume:"))
		self.label.set_padding(3, 3)

		self.vbox.pack_start(self.label, False)
		
		# VU Meter
		self.vu = VUWidget(self, self.mainview)
		self.vbox.pack_start(self.vu, True, True)
				
		self.vbox.show_all()
		self.show_all()
		
	#_____________________________________________________________________
	
	def EmitMinimise(self, widget):
		"""
		Passes the EmitMinimise call coming from an Instrument MixerStrip to GTK.
		
		Parameters:
			widget -- reserved for GTK callbacks, don't use it explicitly.
		"""
		self.emit("minimise")
	
	#_____________________________________________________________________

	def GetLevel(self):
		"""
		Obtain the master level.
		
		Returns:
			the master level value.
		"""
		return self.project.level

	#_____________________________________________________________________

	def GetVolume(self):
		"""
		Obtain the master volume.
		
		Returns:
			the master volume level.
		"""
		return self.project.volume
		
	#_____________________________________________________________________

	def SetVolume(self, vol):
		"""
		Sets the master volume.
		
		Parameters:
			vol -- volume value to set the master volume to.
		"""
		self.project.SetVolume(vol)
		
	#_____________________________________________________________________
	
	def CommitVolume(self):
		#there is no incremental save for project volume
		#as opposed to instrument volume. So do nothing for now.
		pass
	
	#_____________________________________________________________________
	
#=========================================================================



#
#	THIS FILE IS PART OF THE JOKOSHER PROJECT AND LICENSED UNDER THE GPL. SEE
#	THE 'COPYING' FILE FOR DETAILS
#
#	Globals.py
#	
#	This module contains variable definitions that can be used across the code
#	base and also includes methods for reading and writing these settings to
#	the Jokosher configuration in JOKOSHER_CONFIG_HOME/config.
#
#-------------------------------------------------------------------------------

import ConfigParser
import os
import locale, gettext
import pygtk
pygtk.require("2.0")
import gobject, gtk
import xdg.BaseDirectory
import shutil
import PlatformUtils

import gettext
_ = gettext.gettext

class Settings:
	"""
	Handles loading/saving settings from/to a file on disk.
	"""

	# the different settings in each config block
	general = 	{
				# increment each time there is an incompatible change with the config file
				"version" : "1",
				"recentprojects": "value", 
				"startupaction" : "value",
				"projectfolder" : "",
				"windowheight" : 550,
				"windowwidth" : 900,
				"addinstrumentwindowheight" : 350,
				"addinstrumentwindowwidth" : 300,
				"instrumenteffectwindowheight" : 450,				
				"instrumenteffectwindowwidth" : 650,
				
				}

	recording = {
				"fileformat": "flacenc",
				"file_extension": "flac",
				"samplerate": "0", # zero means, autodetect sample rate (ie use any available)
				"audiosrc" : "gconfaudiosrc",
				"device" : "default"
				}
	# Overwrite with platform specific settings
	recording.update( PlatformUtils.GetRecordingDefaults() )
	
	playback = 	{
				"devicename": "default",
				"device": "default",
				"audiosink":"autoaudiosink"
				}
	# Overwrite with platform specific settings
	playback.update( PlatformUtils.GetPlaybackDefaults() )

	extensions = {
				 "extensions_blacklist": ""
				 }
				 
	sections = {
				"General" : general,
				"Recording" : recording,
				"Playback" : playback,
				"Extensions" : extensions
				}

	#_____________________________________________________________________
	
	def __init__(self):
		self.filename = os.path.join(JOKOSHER_CONFIG_HOME, "config")
		
		self.config = ConfigParser.ConfigParser()
		self.read()

	#_____________________________________________________________________

	def read(self):
		"""
		Reads configuration settings from the config file and loads
		then into the Settings dictionaries.
		"""
		self.config.read(self.filename)
	
		for section in self.sections:
			if not self.config.has_section(section):
				self.config.add_section(section)
	
		for section, section_dict in self.sections.iteritems():
			for key, value in self.config.items(section):
				if value == "None":
					value = None
				section_dict[key] = value
	
	#_____________________________________________________________________
		
	def write(self):
		"""
		Writes configuration settings to the Settings config file.
		"""

		for section, section_dict in self.sections.iteritems():
			for key, value in section_dict.iteritems():
				self.config.set(section, key, value)
		
		file = open(self.filename, 'w')
		self.config.write(file)
		file.close()
		
		#_____________________________________________________________________
		
def debug(*listToPrint):
	"""
	Global debug function to redirect all the debugging output from the other
	methods.
	
	Parameters:
		*listToPrint -- list of elements to append to the debugging output.
	"""
	#HACK: we can't import gst at the top of Globals.py because
	#if we do, gstreamer will get to the sys.args and print it's own
	#message instead of ours. This will be fixed once we can use
	#GOption when we depend on pygobject 2.12.
	import gst
	
	message = " ".join( [ str(x) for x in listToPrint ] )
	
	if DEBUG_STDOUT:
		print(message)
	if DEBUG_GST:
		gst.debug(message)
		
#_____________________________________________________________________

def FAT32SafeFilename(filename):
	"""
	Returns a copy fo the given string that has all the
	characters that are not allowed in FAT32 path names
	taken out.
	
	Parameters:
		filename -- the filename string.
	"""
	
	allowedChars = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789$%'`-@{}~!#()&_^ "
	return "".join([x for x in filename if x in allowedChars])

#_____________________________________________________________________

#static list of all the Instrument files (to prevent having to reimport files).
instrumentPropertyList = []
_alreadyCached = False
_cacheGeneratorObject = None

def _cacheInstrumentsGenerator(alreadyLoadedTypes=[]):
	"""
	Yields a loaded Instrument everytime this method is called,
	so that the gui isn't blocked while loading many Instruments.
	If an Instrument's type is already in alreadyLoadedTypes,
	it is considered a duplicate and it's not loaded.
	
	Parameters:
		alreadyLoadedTypes -- array containing the already loaded Instrument types.
		
	Returns:
		the loaded Instrument. *CHECK*
	"""	
	try:
		#getlocale() will usually return  a tuple like: ('en_GB', 'UTF-8')
		lang = locale.getlocale()[0]
	except:
		lang = None
	for instr_path in INSTR_PATHS:
		if not os.path.exists(instr_path):
			continue
		instrFiles = [x for x in os.listdir(instr_path) if x.endswith(".instr")]
		for f in instrFiles:
			config = ConfigParser.SafeConfigParser()
			try:
				config.read(os.path.join(instr_path, f))
			except ConfigParser.MissingSectionHeaderError,e:
				debug("Instrument file %s in %s is corrupt or invalid, not loading"%(f,instr_path))
				continue	

			if config.has_option('core', 'type') and config.has_option('core', 'icon'):
				icon = config.get('core', 'icon')
				type = config.get('core', 'type')
			else:
				continue
			#don't load duplicate instruments
			if type in alreadyLoadedTypes:
				continue
		
			if lang and config.has_option('i18n', lang):
				name = config.get('i18n', lang)
			elif lang and config.has_option('i18n', lang.split("_")[0]):
				#in case lang was 'de_DE', use only 'de'
				name = config.get('i18n', lang.split("_")[0])
			elif config.has_option('i18n', 'en'):
				#fall back on english (or a PO translation, if there is any)
				name = _(config.get( 'i18n', 'en'))
			else:
				continue
			name = unicode(name, "UTF-8")
			pixbufPath = os.path.join(instr_path, "images", icon)
			pixbuf = gtk.gdk.pixbuf_new_from_file(pixbufPath)
			
			# add instrument to defaults list if it's a defaults
			if instr_path == INSTR_PATHS[0]:
				DEFAULT_INSTRUMENTS.append(type)
				
			yield (name, type, pixbuf, pixbufPath)

#_____________________________________________________________________

def getCachedInstruments(checkForNew=False):
	"""
	Creates the Instrument cache if it hasn't been created already and
	return it.
	
	Parameters:
		checkForNew --	True = scan the Instrument folders for new_dir.
						False = don't scan for new Instruments.
						
	Returns:
		a list with the Instruments cached in memory.
	"""
	global instrumentPropertyList, _alreadyCached
	if _alreadyCached and not checkForNew:
		return instrumentPropertyList
	else:
		_alreadyCached = True
	
	listOfTypes = [x[1] for x in instrumentPropertyList]
	try:
		newlyCached = list(_cacheInstrumentsGenerator(listOfTypes))
		#extend the list so we don't overwrite the already cached instruments
		instrumentPropertyList.extend(newlyCached)
	except StopIteration:
		pass

	#sort the instruments alphabetically
	#using the lowercase of the name (at index 0)
	instrumentPropertyList.sort(key=lambda x: x[0].lower())
	return instrumentPropertyList

#_____________________________________________________________________

def getCachedInstrumentPixbuf(get_type):
	for (name, type, pixbuf, pixbufPath) in getCachedInstruments():
		if type == get_type:
			return pixbuf
	return None

#_____________________________________________________________________

def idleCacheInstruments():
	"""
	Loads the Instruments 'lazily' to avoid blocking the GUI.
	
	Returns:
		True -- keep calling itself to load more Instruments.
		False -- stop calling itself and sort Instruments alphabetically.
	"""
	global instrumentPropertyList, _alreadyCached, _cacheGeneratorObject
	if _alreadyCached:
		#Stop idle_add from calling us again
		return False
	#create the generator if it hasnt been already
	if not _cacheGeneratorObject:
		_cacheGeneratorObject = _cacheInstrumentsGenerator()
	
	try:
		instrumentPropertyList.append(_cacheGeneratorObject.next())
		#Make sure idle add calls us again
		return True
	except StopIteration:
		_alreadyCached = True
	
	#sort the instruments alphabetically
	#using the lowercase of the name (at index 0)
	instrumentPropertyList.sort(key=lambda x: x[0].lower())
	#Stop idle_add from calling us again
	return False

#_____________________________________________________________________
	
def PopulateEncoders():
	"""
	Check if the hardcoded list of encoders is available on the system.
	"""
	for type in _export_formats:
		if VerifyAllElements(type[2]):
			#create a dictionary using _export_template as the keys
			#and the current item from _export_formats as the values.
			d = dict(zip(_export_template, type))
			EXPORT_FORMATS.append(d)

#_____________________________________________________________________

def VerifyAllElements(bin_desc):
	#HACK: we can't import gst at the top of Globals.py because
	#if we do, gstreamer will get to the sys.args and print it's own
	#message instead of ours. This will be fixed once we can use
	#GOption when we depend on pygobject 2.12.
	import gst
	
	all_elements_exist = True
	for element in bin_desc.split("!"):
		exists = gst.default_registry_check_feature_version(element.strip(), 0, 10, 0)
		if not exists:
			all_elements_exist = False
			debug('Cannot find "%s" plugin, disabling: "%s"' % (element.strip(), bin_desc))
			# we know at least one of the elements doesnt exist, so skip this encode format.
			break
	
	return all_elements_exist
	
#_____________________________________________________________________

def PopulateAudioBackends():
	CheckBackendList(PLAYBACK_BACKENDS)
	CheckBackendList(CAPTURE_BACKENDS)
	
#_____________________________________________________________________
	
def CheckBackendList(backend_list):
	remove_list = []
	for tuple_ in backend_list:
		bin_desc = tuple_[1]
		if not VerifyAllElements(bin_desc):
			remove_list.append(tuple_)
	
	for tuple_ in remove_list:
		backend_list.remove(tuple_)

#_____________________________________________________________________

def CopyAllFiles(src_dir, dest_dir, only_these_files=None):
	""" Copies all the files, but only the files from one directory to another."""
	for file in os.listdir(src_dir):
		if only_these_files is not None and file not in only_these_files:
			continue
		
		src_path = os.path.join(src_dir, file)
		dest_path = os.path.join(dest_dir, file)
		if os.path.isfile(src_path):
			try:
				shutil.copy2(src_path, dest_path)
			except IOError:
				print "Unable to copy from old ~/.jokosher directory:", src_path

#_____________________________________________________________________

"""
Used for launching the correct help file:
	True -- Jokosher's running locally by the user. Use the help file from
			the help subdirectory.
	False -- Jokosher has been installed system-wide. Use yelp's automatic
			help file selection.
"""
USE_LOCAL_HELP = False

"""
Global paths, so all methods can access them.
If JOKOSHER_DATA_PATH is not set, that is, Jokosher is running locally,
use paths relative to the current running directory instead of /usr ones.
"""

XDG_RESOURCE_NAME = "jokosher"
JOKOSHER_CONFIG_HOME = xdg.BaseDirectory.save_config_path(XDG_RESOURCE_NAME)
JOKOSHER_DATA_HOME =   xdg.BaseDirectory.save_data_path(XDG_RESOURCE_NAME)

data_path = os.getenv("JOKOSHER_DATA_PATH")
if data_path:
	INSTR_PATHS = (os.path.join(data_path, "Instruments"), os.path.join(JOKOSHER_DATA_HOME, "instruments"))
	EXTENSION_PATHS = (os.path.join(data_path, "extensions"), os.path.join(JOKOSHER_DATA_HOME, "extensions"))
	GLADE_PATH = os.path.join(data_path, "Jokosher.glade")
else:
	data_path = os.path.dirname(os.path.abspath(__file__))
	INSTR_PATHS = (os.path.join(data_path, "..", "Instruments"), os.path.join(JOKOSHER_DATA_HOME, "instruments"))
	EXTENSION_PATHS = (os.path.join(data_path, "..", "extensions"), os.path.join(JOKOSHER_DATA_HOME, "extensions"))
	GLADE_PATH = os.path.join(data_path, "Jokosher.glade")
	LOCALE_PATH = os.path.join(data_path, "..", "locale")

# create a couple dirs to avoid having problems creating a non-existing
# directory inside another non-existing directory
create_dirs = [
	'extensions',
	'instruments',
	('instruments', 'images'),
	'presets',
	('presets', 'effects'),
	('presets', 'mixdown'),
	'mixdownprofiles',
	'templates',
]

# do a listing before we create the dirs so we know if it was empty (ie first run)
jokosher_dir_empty = (len(os.listdir(JOKOSHER_DATA_HOME)) == 0)
_HOME_DOT_JOKOSHER = os.path.expanduser("~/.jokosher")

if jokosher_dir_empty and os.path.isdir(_HOME_DOT_JOKOSHER):
	# Copying old config file from ~/.jokosher.
	CopyAllFiles(_HOME_DOT_JOKOSHER, JOKOSHER_CONFIG_HOME, ["config"])

for dirs in create_dirs:
	if isinstance(dirs, str):
		new_dir = os.path.join(JOKOSHER_DATA_HOME, dirs)
		old_dir = os.path.join(_HOME_DOT_JOKOSHER, dirs)
	else:
		new_dir = os.path.join(JOKOSHER_DATA_HOME, *dirs)
		old_dir = os.path.join(_HOME_DOT_JOKOSHER, *dirs)
		
	if not os.path.isdir(new_dir):
		try:
			os.makedirs(new_dir)
		except:
			raise "Failed to create user config directory %s" % new_dir
	
	if jokosher_dir_empty and os.path.isdir(old_dir) and os.path.isdir(new_dir):
		CopyAllFiles(old_dir, new_dir)
			

#TODO: make this a list with the system path and home directory path
EFFECT_PRESETS_PATH = os.path.join(JOKOSHER_DATA_HOME, "presets", "effects")
TEMPLATES_PATH = os.path.join(JOKOSHER_DATA_HOME, "templates")
MIXDOWN_PROFILES_PATH = os.path.join(JOKOSHER_DATA_HOME, "mixdownprofiles")

IMAGE_PATH = os.getenv("JOKOSHER_IMAGE_PATH")
if not IMAGE_PATH:
	IMAGE_PATH = os.path.join(data_path, "..", "images")

LOCALE_PATH = os.getenv("JOKOSHER_LOCALE_PATH")
if not LOCALE_PATH:
	LOCALE_PATH = os.path.join(data_path, "..", "locale")

HELP_PATH = os.getenv("JOKOSHER_HELP_PATH")
if not HELP_PATH:
	USE_LOCAL_HELP = True
	
	# change the local help file to match the current locale
	current_locale = "C"
	if locale.getlocale()[0] and not locale.getlocale()[0].startswith("en", 0, 2):
		current_locale = locale.getlocale()[0][:2]
		
	HELP_PATH = os.path.join(data_path, "..", "help/jokosher",
							 current_locale, "jokosher.xml")
	
	# use C (en) as the default help fallback
	if not os.path.exists(HELP_PATH):
		HELP_PATH = os.path.join(data_path, "..", "help/jokosher/C/jokosher.xml")

# add your own extension dirs with envar JOKOSHER_EXTENSION_DIRS, colon-separated
__extra_ext_dirs = os.environ.get('JOKOSHER_EXTENSION_DIRS','')
if __extra_ext_dirs:
	EXTENSION_PATHS = __extra_ext_dirs.split(':') + list(EXTENSION_PATHS)

""" ExtensionManager data """
AVAILABLE_EXTENSIONS = []

""" Locale constant """
LOCALE_APP = "jokosher"

""" Categories enum """
class Categories:
	(broken, unclassified, amplifiers, chorus, compressors,
	delays, distortions, equalizers, filters, flangers,
	miscellaneous, modulators, oscillators, phasers, reverbs,
	simulators) = range(16)

""" Set in Project.py """
VERSION = None
EFFECT_PRESETS_VERSION = None
LADSPA_FACTORY_REGISTRY = None
LADSPA_NAME_MAP = []
LADPSA_CATEGORIES_LIST = [
						(_("Broken"), "effect_broken.png"),
						(_("Unclassified"), "effect_unclassified.png"),
						(_("Amplifiers"), "effect_amplifiers.png"),
						(_("Chorus"), "effect_chorus.png"),
						(_("Compressors"), "effect_compressors.png"),
						(_("Delays"), "effect_delays.png"),
						(_("Distortions"), "effect_distortion.png"),
						(_("Equalizers"), "effect_equalizers.png"),
						(_("Filters"), "effect_filters.png"),
						(_("Flangers"), "effect_flangers.png"),
						(_("Miscellaneous"), "effect_miscellaneous.png"),
						(_("Modulators"), "effect_modulators.png"),
						(_("Oscillators"), "effect_oscillators.png"),
						(_("Phasers"), "effect_phasers.png"),
						(_("Reverbs"), "effect_reverbs.png"),
						(_("Simulators"), "effect_simulators.png")
						]
LADSPA_CATEGORIES_DICT = {
						"ladspa-SweepVFII" : Categories.modulators,
						"ladspa-SweepVFI" : Categories.modulators,
						"ladspa-PhaserII" : Categories.phasers,
						"ladspa-PhaserI" : Categories.phasers,
						"ladspa-ChorusII" : Categories.chorus,
						"ladspa-ChorusI" : Categories.chorus,
						"ladspa-Clip" : Categories.amplifiers,
						"ladspa-CabinetII" : Categories.simulators,
						"ladspa-CabinetI" : Categories.simulators,
						"ladspa-AmpV" : Categories.simulators,
						"ladspa-AmpIV" : Categories.simulators,
						"ladspa-AmpIII" : Categories.simulators,
						"ladspa-PreampIV" : Categories.simulators,
						"ladspa-PreampIII" : Categories.simulators,
						"ladspa-Compress" : Categories.compressors,
						"ladspa-Eq" : Categories.equalizers,
						"ladspa-ssm-masher" : Categories.broken, #no sound
						"ladspa-slew-limiter-rc" : Categories.broken, #no sound
						"ladspa-slide-tc" : Categories.broken, #chirps then dies
						"ladspa-signal-abs-cr" : Categories.modulators,
						"ladspa-vcf-hshelf" : Categories.broken, #erratic behavior.
						"ladspa-vcf-lshelf" : Categories.broken, #erratic behavior
						"ladspa-vcf-peakeq" : Categories.filters,
						"ladspa-vcf-notch" : Categories.filters,
						"ladspa-vcf-bp2" : Categories.filters,
						"ladspa-vcf-bp1" : Categories.broken, #no sound
						"ladspa-vcf-hp" : Categories.filters,
						"ladspa-vcf-lp" : Categories.filters,
						"ladspa-vcf-reslp" : Categories.filters,
						"ladspa-range-trans-cr" : Categories.amplifiers, #works, but the settings are impossible to use properly
						"ladspa-hz-voct-ar" : Categories.broken, #no sound
						"ladspa-Phaser1+LFO" : Categories.phasers,
						"ladspa-Chorus2" : Categories.chorus, #so so
						"ladspa-Chorus1" : Categories.chorus, # so so
						"ladspa-tap-vibrato" : Categories.modulators,
						"ladspa-tap-tubewarmth" : Categories.filters,
						"ladspa-tap-tremolo" : Categories.modulators,
						"ladspa-tap-sigmoid" : Categories.amplifiers,
						"ladspa-tap-reflector" : Categories.modulators,
						"ladspa-tap-pitch" : Categories.modulators,
						"ladspa-tap-pinknoise" : Categories.miscellaneous,
						"ladspa-tap-limiter" : Categories.amplifiers,
						"ladspa-tap-equalizer-bw" : Categories.equalizers,
						"ladspa-tap-equalizer" : Categories.equalizers,
						"ladspa-formant-vc" : Categories.modulators,
						"ladspa-tap-deesser" : Categories.filters,
						"ladspa-tap-dynamics-m" : Categories.filters, #could be in another category
						"ladspa-imp" : Categories.filters,
						"ladspa-pitchScaleHQ" : Categories.modulators, #crap
						"ladspa-mbeq" : Categories.equalizers,
						"ladspa-sc4m" : Categories.filters, #could be in another category
						"ladspa-artificialLatency" : Categories.miscellaneous,
						"ladspa-pitchScale" : Categories.modulators, #crap
						"ladspa-pointerCastDistortion" : Categories.distortions, #crap
						"ladspa-const" : Categories.distortions, #could be in another category
						"ladspa-lsFilter" : Categories.filters,
						"ladspa-revdelay" : Categories.delays,
						"ladspa-delay-c" : Categories.broken, #erratic behavior
						"ladspa-delay-l" : Categories.broken, #no change in sound?
						"ladspa-delay-n" : Categories.broken, #no change in sound?
						"ladspa-decay" : Categories.distortions, #controls make it unusable
						"ladspa-comb-c" : Categories.broken, #erratic behavior
						"ladspa-comb-l" : Categories.broken, #no change in sound?
						"ladspa-comb-n" : Categories.broken, #no change in sound and static
						"ladspa-allpass-c" : Categories.broken, #no change in sound?
						"ladspa-allpass-l" : Categories.broken, #no change in sound?
						"ladspa-allpass-n" : Categories.broken, #no change in sound?
						"ladspa-butthigh-iir" : Categories.filters,
						"ladspa-buttlow-iir" : Categories.filters,
						"ladspa-dj-eq-mono" : Categories.equalizers,
						"ladspa-notch-iir" : Categories.filters,
						"ladspa-lowpass-iir" : Categories.filters,
						"ladspa-highpass-iir" : Categories.filters,
						"ladspa-bandpass-iir" : Categories.filters,
						"ladspa-bandpass-a-iir" : Categories.filters,
						"ladspa-gongBeater" : Categories.modulators, #crap
						"ladspa-djFlanger" : Categories.flangers,
						"ladspa-giantFlange" : Categories.flangers,
						"ladspa-amPitchshift" : Categories.modulators,
						"ladspa-chebstortion" : Categories.distortions, #weak
						"ladspa-inv" : Categories.broken, #no change in sound, no options either
						"ladspa-zm1" : Categories.broken, #no change in sound, no options either
						"ladspa-sc1" : Categories.compressors, #could be in another category
						"ladspa-gong" : Categories.filters,
						"ladspa-freqTracker" : Categories.broken, #no sound
						"ladspa-rateShifter" : Categories.filters,
						"ladspa-fmOsc" : Categories.broken, #erratic behavior
						"ladspa-smoothDecimate" : Categories.filters,
						"ladspa-hardLimiter" : Categories.amplifiers,
						"ladspa-gate" : Categories.filters, #could be in another category
						"ladspa-satanMaximiser" : Categories.distortions,
						"ladspa-alias" : Categories.filters, #could be in another category
						"ladspa-valveRect" : Categories.filters,
						"ladspa-crossoverDist" : Categories.distortions, #crap
						"ladspa-dysonCompress" : Categories.compressors,
						"ladspa-delayorama" : Categories.delays,
						"ladspa-autoPhaser" : Categories.phasers,
						"ladspa-fourByFourPole" : Categories.filters,
						"ladspa-lfoPhaser" : Categories.phasers,
						"ladspa-gsm" : Categories.modulators,
						"ladspa-svf" : Categories.filters,
						"ladspa-foldover" : Categories.distortions,
						"ladspa-harmonicGen" : Categories.modulators, #crap
						"ladspa-sifter" : Categories.modulators, #sounds like Distortion
						"ladspa-valve" : Categories.distortions, #weak
						"ladspa-tapeDelay" : Categories.delays,
						"ladspa-dcRemove" : Categories.broken, #no change in sound, no options either
						"ladspa-fadDelay" : Categories.delays, #psychedelic stuff
						"ladspa-transient" : Categories.modulators,
						"ladspa-triplePara" : Categories.filters,
						"ladspa-singlePara" : Categories.filters,
						"ladspa-retroFlange" : Categories.flangers,
						"ladspa-flanger" : Categories.flangers,
						"ladspa-decimator" : Categories.filters,
						"ladspa-hermesFilter" : Categories.filters, #control needs to have 2 columns, doesn't fit screen
						"ladspa-multivoiceChorus" : Categories.chorus,
						"ladspa-foverdrive" : Categories.distortions,
						"ladspa-declip" : Categories.filters, #couldn't properly test it since I had no clipping audio
						"ladspa-comb" : Categories.filters,
						"ladspa-ringmod-1i1o1l" : Categories.modulators,
						"ladspa-shaper" : Categories.filters,
						"ladspa-divider" : Categories.filters,
						"ladspa-diode" : Categories.distortions,
						"ladspa-amp" : Categories.amplifiers,
						"ladspa-Parametric1" : Categories.filters,
						"ladspa-wshape-sine" : Categories.broken, #no change in sound?
						"ladspa-vcf303" : Categories.filters,
						"ladspa-limit-rms" : Categories.broken, #controls make it unusable
						"ladspa-limit-peak" : Categories.broken, #controls make it unusable
						"ladspa-expand-rms" : Categories.broken, #controls make it unusable
						"ladspa-expand-peak" : Categories.broken, #controls make it unusable
						"ladspa-compress-rms" : Categories.broken, #controls make it unusable
						"ladspa-compress-peak" : Categories.broken, #controls make it unusable
						"ladspa-identity-audio" : Categories.broken, #no change in sound?
						"ladspa-hard-gate" : Categories.filters,
						"ladspa-grain-scatter" : Categories.broken, #no sound
						"ladspa-fbdelay-60s" : Categories.delays,
						"ladspa-fbdelay-5s" : Categories.delays,
						"ladspa-fbdelay-1s" : Categories.delays,
						"ladspa-fbdelay-0-1s" : Categories.delays,
						"ladspa-fbdelay-0-01s" : Categories.delays,
						"ladspa-delay-60s" : Categories.delays,
						"ladspa-delay-1s" : Categories.delays,
						"ladspa-delay-0-1s" : Categories.delays,
						"ladspa-delay-0-01s" : Categories.delays,
						"ladspa-disintegrator" : Categories.filters, #crap
						"ladspa-triangle-fcsa-oa" : Categories.oscillators,
						"ladspa-triangle-fasc-oa" : Categories.broken, #no sound
						"ladspa-syncsquare-fcga-oa" : Categories.oscillators,
						"ladspa-syncpulse-fcpcga-oa" : Categories.oscillators,
						"ladspa-sum-iaic-oa" : Categories.filters,
						"ladspa-square-fa-oa" : Categories.oscillators,
						"ladspa-sinusWavewrapper" : Categories.filters,
						"ladspa-ratio-ncda-oa" : Categories.distortions,
						"ladspa-ratio-nadc-oa" : Categories.broken, #no sound
						"ladspa-random-fcsa-oa" : Categories.oscillators, #we GOTTA call this Atari or Arcade. It's the same sound!
						"ladspa-random-fasc-oa" : Categories.broken, #no sound
						"ladspa-sawtooth-fa-oa" : Categories.oscillators,
						"ladspa-pulse-fcpa-oa" : Categories.oscillators,
						"ladspa-pulse-fapc-oa" : Categories.oscillators,
						"ladspa-product-iaic-oa" : Categories.oscillators,
						"ladspa-lp4pole-fcrcia-oa" : Categories.filters,
						"ladspa-fmod-fcma-oa" : Categories.filters,
						"ladspa-fmod-famc-oa" : Categories.broken, #controls make it unusable
						"ladspa-amp-gcia-oa" : Categories.broken, #controls make it unusable
						"ladspa-difference-icma-oa" : Categories.amplifiers,
						"ladspa-difference-iamc-oa" : Categories.broken, #no sound
						"ladspa-sine-fcaa" : Categories.oscillators,
						"ladspa-sine-faac" : Categories.broken, #no sound
						"ladspa-hpf" : Categories.filters,
						"ladspa-lpf" : Categories.filters,
						"ladspa-adsr" : Categories.broken, #controls make it unusable, no sound
						"ladspa-amp-mono" : Categories.amplifiers,
						"ladspa-delay-5s" : Categories.delays
						}
DEBUG_STDOUT, DEBUG_GST = (False, False)

_export_template = ("description", "extension", "pipeline") 
_export_formats = 	[
					("Ogg Vorbis", "ogg", "vorbisenc ! oggmux"),
					("MP3", "mp3", "lame"),
					("Flac", "flac", "flacenc"),
					("WAV", "wav", "wavenc"),
					]

EXPORT_FORMATS = []

SAMPLE_RATES = [8000, 11025, 22050, 32000, 44100, 48000, 96000, 192000]


PLAYBACK_BACKENDS = [
	(_("Autodetect"), "autoaudiosink"),
	(_("Use GNOME Settings"), "gconfaudiosink"),
	("ALSA", "alsasink"),
	("OSS", "osssink"),
	("JACK", "jackaudiosink"),
	("PulseAudio", "pulsesink"),
	("Direct Sound", "directsoundsink"),
	("Core Audio", "osxaudiosink")
]

CAPTURE_BACKENDS = [
	(_("GNOME Settings"), "gconfaudiosrc"),
	("ALSA", "alsasrc"),
	("OSS", "osssrc"),
	("JACK", "jackaudiosrc"),
	("PulseAudio", "pulsesrc"),
	("Direct Sound", "dshowaudiosrc"),
	("Core Audio", "osxaudiosrc")
]

""" Default Instruments """
DEFAULT_INSTRUMENTS = []
""" init Settings """
settings = Settings()

""" Cache Instruments """
gobject.idle_add(idleCacheInstruments)


gobject.set_application_name(_("Jokosher Audio Editor"))
gobject.set_prgname(LOCALE_APP)
gtk.window_set_default_icon_name("jokosher")
# environment variable for pulseaudio type
os.environ["PULSE_PROP_media.role"] = "production"

# I have decided that Globals.py is a boring source file. So, here is a little
# joke. What does the tax office and a pelican have in common? They can both stick
# their bills up their arses. Har har har.



#	GNOMEAudioProfiles.py
#	--------------------
#	This extension is for adding the GNOME Audio Profiles into Jokosher

import Jokosher.Extension
import gconf
import traceback

EXTENSION_NAME = "GNOME Audio Profiles"
EXTENSION_DESCRIPTION = "Imports GNOME Audio Profiles into jokosher"
EXTENSION_VERSION = "0.0.1"
_GCONF_PROFILE_PATH = "/system/gstreamer/0.10/audio/profiles/"
_GCONF_PROFILE_LIST_PATH = "/system/gstreamer/0.10/audio/global/profile_list"
audio_profiles_list = []
API = None

def startup(api):
	global API, audio_profiles_list
	API = api
	_GCONF = gconf.client_get_default()
	profiles = _GCONF.get_list(_GCONF_PROFILE_LIST_PATH, 1)
	for name in profiles:
		if (_GCONF.get_bool(_GCONF_PROFILE_PATH + name + "/active")):
			description = _GCONF.get_string(_GCONF_PROFILE_PATH + name + "/name")
			extension = _GCONF.get_string(_GCONF_PROFILE_PATH + name + "/extension")
			encodeBin = _GCONF.get_string(_GCONF_PROFILE_PATH + name + "/pipeline")
			
			#last parameter is False to tell Jokosher to assume the strings are correct, and not check them.
			error = API.add_export_format(description, extension, encodeBin, False)
			if error == 0:
				#it has been succesfully added, so keep track of what we added
				audio_profiles_list.append((description, extension, encodeBin))

def shutdown():
	for format in audio_profiles_list:
		API.remove_export_format(*format)



#
#	THIS FILE IS PART OF THE JOKOSHER PROJECT AND LICENSED UNDER THE GPL. SEE
#	THE 'COPYING' FILE FOR DETAILS
#
#	InstrumentEffectsDialog.py
#	
#	This module us used to present the dialog for adding and managing instrument
#	effects. Note that this module deals with the main effects dialog *and* the
#	effect settings dialog, which we're both handed to Mr. Bacon in a spec by the
#	Gods of Metal. They came to him in a dream to warn against a bleak future where
#	massively distorted music could not be made using linux.
#
#-------------------------------------------------------------------------------

import gtk.glade
import gobject
import cairo
import pygst
pygst.require("0.10")
import gst
import os
import Globals
import EffectPresets

import gettext
_ = gettext.gettext

#=========================================================================

class InstrumentEffectsDialog:
	"""
	This class displays and implements the Instrument effects dialog
	box, which can be used to add, remove, and edit effects for an
	Instrument.
	"""
	
	#_____________________________________________________________________	
	
	def __init__(self, instrument, destroyCallback, windowIcon):
		"""
		This constructor enables a lot of variables, reads in the glade
		file for the main dialog, and populates the effects and presets
		controls.
		
		Parameters:
			instrument -- Instrument whose effects are being modified.
			destroyCallback -- GTK callback. Called when this dialog gets destroyed.
			windowIcon -- icon to use on the window's upper corner.
		"""
		# a reference to the instrument object
		self.instrument = instrument
		self.windowIcon = windowIcon
		self.res = gtk.glade.XML(Globals.GLADE_PATH, "InstrumentEffectsDialog")

		self.Updating = False
		self.effectWidgets = []
		
		self.signals = {
			"on_comboPresets_changed" : self.OnEffectChainPresetChanged,
			"on_buttonPresetSave_clicked" : self.OnEffectChainPresetSave,
			"on_buttonPresetDelete_clicked" : self.OnEffectChainPresetDelete,
			
			"on_comboCategories_changed" : self.OnCategoryChanged,
			"on_listEffects_row_activated" : self.OnEffectActivated,
			"on_listActiveEffects_row_activated" : self.OnActiveEffectActivated,
			"on_buttonEffectAdd_clicked" : self.OnEffectAdd,
			"on_buttonEffectUp_clicked" : self.OnEffectUp,
			"on_buttonEffectDown_clicked" : self.OnEffectDown,
			"on_buttonEffectDelete_clicked" : self.OnEffectDeleted,
			"on_buttonEffectSettings_clicked" : self.OnEffectSettings,
			
			"on_buttonPlay_clicked" : self.OnPlay,
			"on_buttonClose_clicked" : self.OnClose,
			"on_InstrumentEffects_configure_event" : self.OnResize,
			"on_InstrumentEffects_destroy" : self.OnDestroy
		}
		
		# auto connect the signals to the methods
		self.res.signal_autoconnect(self.signals)

		# grab references to some Glade items
		self.window = self.res.get_widget("InstrumentEffectsDialog")
		self.mainvbox = self.res.get_widget("InstrumentEffectsDialogVBox")
		
		self.imageInstrument = self.res.get_widget("imageInstrument")
		self.comboPresets = self.res.get_widget("comboPresets")
		self.buttonPresetSave = self.res.get_widget("buttonPresetSave")
		self.buttonPresetDelete = self.res.get_widget("buttonPresetDelete")
		
		self.comboCategories = self.res.get_widget("comboCategories")
		self.listEffects = self.res.get_widget("listEffects")
		self.buttonEffectAdd = self.res.get_widget("buttonEffectAdd")
		self.labelActiveEffects = self.res.get_widget("labelActiveEffects")
		self.listActiveEffects = self.res.get_widget("listActiveEffects")
		self.buttonEffectUp = self.res.get_widget("buttonEffectUp")
		self.buttonEffectDown = self.res.get_widget("buttonEffectDown")
		self.buttonEffectDelete = self.res.get_widget("buttonEffectDelete")
		self.buttonEffectSettings = self.res.get_widget("buttonEffectSettings")

		self.buttonPlay = self.res.get_widget("buttonPlay")
		self.buttonClose = self.res.get_widget("buttonClose")

		# connect the right-click signal for both treeviews
		self.listEffects.connect("button-press-event", self.OnEffectsTreeViewClick)
		self.listActiveEffects.connect("button-press-event", self.OnActiveEffectsTreeViewClick)

		# connect the destroy signal and set the window icon
		self.window.connect("destroy", destroyCallback)
		self.window.set_icon(self.windowIcon)
		
		# set single selection for the list views
		self.listEffects.get_selection().set_mode(gtk.SELECTION_SINGLE)
		self.listActiveEffects.get_selection().set_mode(gtk.SELECTION_SINGLE)

		# create and pack the header image
		self.headerCairoImage = CairoDialogHeaderImage(_("Instrument Effects"))
		self.headerCairoImage.set_size_request(self.window.allocation.width, 60)
		self.mainvbox.pack_start(self.headerCairoImage, False, False)
		self.headerCairoImage.show()
		
		# set the instrument name and image from self.instrument
		self.imageInstrument.set_from_pixbuf(self.instrument.pixbuf.scale_simple(32, 32, gtk.gdk.INTERP_BILINEAR))
		self.labelActiveEffects.set_label(self.labelActiveEffects.get_label()+self.instrument.name)
		
		# create the appropriate models
		self.modelCategories = gtk.ListStore(gtk.gdk.Pixbuf, str) #image, category
		self.modelPresets = gtk.ListStore(str) #name
		self.modelEffects = gtk.ListStore(gtk.gdk.Pixbuf, str, str, str, bool) #image, shortname, longname, category, listed
		self.modelActiveEffects = gtk.ListStore(gtk.gdk.Pixbuf, str, str, int) #image, shortname, longname, uniqueID

		# create a model filter for the effects model, and set it's visible column
		self.filterEffects = self.modelEffects.filter_new()
		self.filterEffects.set_visible_column(4)

		# set the models for each widget accordingly
		self.comboCategories.set_model(self.modelCategories)
		self.comboPresets.set_model(self.modelPresets)
		self.listEffects.set_model(self.filterEffects)
		self.listActiveEffects.set_model(self.modelActiveEffects)
		
		# create the columns with their respective renderers and add them	
		self.listEffects.append_column(gtk.TreeViewColumn(_("Name"), gtk.CellRendererText(), text=2))
		self.listActiveEffects.append_column(gtk.TreeViewColumn(_("Type"), gtk.CellRendererPixbuf(), pixbuf=0))
		self.listActiveEffects.append_column(gtk.TreeViewColumn(_("Name"), gtk.CellRendererText(), text=2))
		
		# effects combo
		pixRenderer = gtk.CellRendererPixbuf()
		pixRenderer.set_property("xpad", 6)
		pixRenderer.set_property("ypad", 3)
		textRenderer = gtk.CellRendererText()
		self.comboCategories.clear()					# all combos by default have one textRenderer. Remove it
														# before adding the pixRenderer.
		self.comboCategories.pack_start(pixRenderer, expand=False)
		self.comboCategories.pack_start(textRenderer)
		self.comboCategories.add_attribute(pixRenderer, "pixbuf", 0)
		self.comboCategories.add_attribute(textRenderer, "text", 1)
		
		# presets combo
		self.comboPresets.clear()
		textRenderer = gtk.CellRendererText()
		self.comboPresets.pack_start(textRenderer)
		self.comboPresets.add_attribute(textRenderer, "text", 0)
		
		# create a copy of the categories list to avoid errors in the for's
		categories = Globals.LADPSA_CATEGORIES_LIST
		
		#the temporary list to sort them
		tempEffectList = []
		# load all available effects into modelEffects
		for effect in Globals.LADSPA_NAME_MAP:
			newEffect = [None, None, None, None, None]

			try:
				imageFile = categories[Globals.LADSPA_CATEGORIES_DICT[effect[0]]][1]	# image path
				category = categories[Globals.LADSPA_CATEGORIES_DICT[effect[0]]][0]
			except KeyError:	# unclassified
				imageFile = categories[1][1]
				category = categories[1][0]
				
			newEffect[0] = gtk.gdk.pixbuf_new_from_file(os.path.join(Globals.IMAGE_PATH, imageFile)) # image
			newEffect[1] = effect[0]	# short name
			newEffect[2] = effect[1]	# long name
			newEffect[3] = category
			newEffect[4] = False		# listed in the left effects pane
			
			tempEffectList.append(newEffect)
		
		tempEffectList.sort(key=lambda x: x[2].lower()) #sort alphabetically by long name
		for newEffect in tempEffectList:
			self.modelEffects.append(newEffect)
		
		# create a list with available presets and populate the model
		self.presets = EffectPresets.EffectPresets()
		self._LoadInstrumentPresets()
				
		# append the categories and their image to the categories combo box
		for index, category in enumerate(categories):
			# skip the Broken category
			if index == 0:
				continue
				
			for effect in self.modelEffects:
				#If there is at least one effect in this category; always include unclassified
				if effect[3] == category[0] or index == 1:	
					imageFile = category[1]
					icon = gtk.gdk.pixbuf_new_from_file(os.path.join(Globals.IMAGE_PATH, imageFile))
					self.modelCategories.append([icon, category[0]])
					break
		
		#set the first one active
		self.comboCategories.set_active(0)
		
		# this says if the project is playing, so we know to toggle the
		# play button in the dialog	
		self.isPlaying = self.instrument.project.GetIsPlaying()

		if self.isPlaying:
			self.buttonPlay.set_use_stock(True)
			self.buttonPlay.set_label(gtk.STOCK_MEDIA_STOP)

		# listen to the Project, Instrument and Preset changes
		self.instrument.project.connect("audio-state::play", self.OnProjectPlay)
		self.instrument.project.connect("audio-state::record", self.OnProjectRecord)
		self.instrument.project.connect("audio-state::stop", self.OnProjectStop)
		self.instrument.connect("effect", self.OnInstrumentEffect)
		self.presets.connect("single-preset", self.OnSinglePreset)
		self.presets.connect("chain-preset", self.OnChainPreset)

		self.updatinggui = True
		if self.instrument.currentchainpreset is not None:
			self.comboPresets.set_active(self.instrument.currentchainpreset)
		self.updatinggui = False

		self.width = int(Globals.settings.general["instrumenteffectwindowwidth"])
		self.height = int(Globals.settings.general["instrumenteffectwindowheight"])
		self.window.resize(self.width, self.height)

			
		self.Update()
			
	#_____________________________________________________________________
	
	def _LoadInstrumentPresets(self):
		"""
		Loads the presets for this Instrument type if there are any available.
		It then adds them to the chain presets combo box model.
		"""
		self.availpresets = []
		self.modelPresets.clear()
		
		self.presets.FillEffectsPresetsRegistry()
		
		try:
			instrPresets = self.presets.effectpresetregistry["instruments"][self.instrument.instrType] 
			for pres in instrPresets:
				if instrPresets[pres]["dependencies"].issubset(Globals.LADSPA_FACTORY_REGISTRY):
					self.availpresets.append(pres)
					self.modelPresets.append([pres])
		except KeyError:
			# no presets for this Instrument
			pass
		
	#_____________________________________________________________________
	
	def Update(self):
		"""
		Refreshes the effects inside the modelActiveEffects when they are
		added, removed or shuffled.
		"""
		if self.Updating:
			return
		self.Updating = True
		
		# create a copy of the categories list to avoid errors in the for
		categories = Globals.LADPSA_CATEGORIES_LIST
		
		# clear the active effects list
		self.modelActiveEffects.clear()
		
		# append all the active effects to the modelActiveEffects
		for effect in self.instrument.effects:
			shortName =  effect.get_factory().get_name()
			longName = effect.get_factory().get_longname()	
			activeEffect = [None, None, None, None]
			
			try:
				imageFile = categories[Globals.LADSPA_CATEGORIES_DICT[shortName]][1]	# image path
			except KeyError:
				imageFile = categories[1][1]	# unclassified
				
			activeEffect[0] = gtk.gdk.pixbuf_new_from_file(os.path.join(Globals.IMAGE_PATH, imageFile))
			activeEffect[1] = shortName
			activeEffect[2] = longName
			activeEffect[3] = len(self.modelActiveEffects)
			
			self.modelActiveEffects.append(activeEffect)
		
		self.Updating = False
		
	#_____________________________________________________________________	
		
	def OnClose(self, button):
		"""
		If the OK button is pressed on the dialog box, the window is destroyed.
		
		Parameters:
			button -- reserved for GTK callbacks, don't use it explicitly.
		"""
		self.instrument.project.disconnect_by_func(self.OnProjectPlay)
		self.instrument.project.disconnect_by_func(self.OnProjectRecord)
		self.instrument.project.disconnect_by_func(self.OnProjectStop)
		self.instrument.disconnect_by_func(self.OnInstrumentEffect)
		self.window.destroy()
		
	#_____________________________________________________________________	

	def OnPlay(self, button):
		"""
		Pressing the Play/Stop button on the dialog box allows the user
		to play back the project to test if the effect settings are
		right for them. When user press the Play button, it switches to
		a stop button, and vice versa.
		
		Parameters:
			button -- reserved for GTK callbacks, don't use it explicitly.
		"""
		if self.isPlaying == False:
			# things to do if the project is not already playing, and hence
			# needs to start playing
			self.instrument.project.Play()
			self.isPlaying = True
		else:
			self.instrument.project.Stop()
			self.isPlaying = False

	#_____________________________________________________________________
	
	def OnSinglePreset(self, effectPresets):
		"""
		Callback function for when the a the list of single effects is modified.
		"""
		self._LoadEffectPresets()
	
	#_____________________________________________________________________
	
	def OnChainPreset(self, effectPresets):
		"""
		Callback function for when the a the list of chain effects is modified.
		"""
		self._LoadInstrumentPresets()
	
	#_____________________________________________________________________
	
	def OnProjectRecord(self, project):
		"""
		Callback for when the project starts recording.
		
		Parameters:
			project -- The project instance that send the signal.
		"""
		self.window.set_sensitive(False)
	
	#_____________________________________________________________________
	
	def OnProjectPlay(self, project):
		"""
		Callback for when the project starts playing.
		
		Parameters:
			project -- The project instance that send the signal.
		"""
		# things to do if the project is not already playing, and hence
		# needs to start playing
		self.buttonPlay.set_use_stock(True)
		self.buttonPlay.set_label(gtk.STOCK_MEDIA_STOP)
		
		# set this to True to show we are now playing
		self.isPlaying = True
	
	#_____________________________________________________________________
	
	def OnProjectStop(self, project):
		"""
		Callback for when the project stops playing or recording.
		
		Parameters:
			project -- The project instance that send the signal.
		"""
		self.window.set_sensitive(True)
		# things to do when the stop button is pressed to stop playback
		self.buttonPlay.set_use_stock(True)
		self.buttonPlay.set_label(gtk.STOCK_MEDIA_PLAY)
		
		# set this to False to show we are no longer playing
		self.isPlaying = False
	
	#_____________________________________________________________________
	
	def OnInstrumentEffect(self, instrument):
		"""
		Callback for when the effects on the instrument change.
		
		Parameters:
			instrument -- the instrument instance that send the signal.
		"""
		self.Update()
	
	#_____________________________________________________________________
	
	def OnCategoryChanged(self, combo):
		"""
		Updates the list of available effects depending on the category chosen.
		
		Parameters:
			combo -- reserved for GTK callbacks, don't use it explicitly.
		"""
		# display all the correspondent effects in the effects list
		for effect in self.modelEffects:
			if effect[3] == self.modelCategories[combo.get_active()][1]:	#match the category
				effect[4] = True	#display the effect
			else:
				effect[4] = False	#hide the effect
	
	#_____________________________________________________________________
	
	def OnActiveEffectsTreeViewClick(self, widget, mouse):
		"""
		Called when the user presses a mouse button over the ListActiveEffects TreeView.
		If it's a right-click, creates a context menu on the fly for
		manipulating effects.
		
		Parameters:
			widget -- reserved for GTK callbacks, don't use it explicitly.
			mouse -- GTK mouse event that fired this method call.
		"""
		selection = self.listActiveEffects.get_selection().get_selected()
		
		# return if there is no active selection
		if not selection[1]:
			return
		
		# Create context menu on a right-click
		if mouse.button == 3:
			menu = gtk.Menu()
			items = [
					(_("_Move up..."), self.OnEffectUp, True, gtk.image_new_from_stock(gtk.STOCK_GO_UP, gtk.ICON_SIZE_MENU)),
					(_("M_ove down..."), self.OnEffectDown, True, gtk.image_new_from_stock(gtk.STOCK_GO_DOWN, gtk.ICON_SIZE_MENU)),
					("---", None, False, None),
					(_("_Delete"), self.OnEffectDeleted, True, gtk.image_new_from_stock(gtk.STOCK_DELETE, gtk.ICON_SIZE_MENU)),
					(_("_Settings"), self.OnEffectSettings, True, gtk.image_new_from_stock(gtk.STOCK_PROPERTIES, gtk.ICON_SIZE_MENU))
					]
			
			for label, callback, sensitivity, image in items:
				if label == "---":
					menuItem = gtk.SeparatorMenuItem()
				elif image:
					menuItem = gtk.ImageMenuItem(label, True)
					menuItem.set_image(image)
				else:
					menuItem = gtk.MenuItem(label=label)
				
				menuItem.set_sensitive(sensitivity)
				menuItem.show()
				menu.append(menuItem)
				if callback:
					menuItem.connect("activate", callback)
			
			menu.popup(None, None, None, mouse.button, mouse.time)
	#_____________________________________________________________________
	
	def OnEffectsTreeViewClick(self, widget, mouse):
		"""
		Called when the user presses a mouse button over the ListEffects TreeView.
		If it's a right-click, creates a context menu on the fly for
		manipulating effects.
		
		Parameters:
			widget -- reserved for GTK callbacks, don't use it explicitly.
			mouse -- GTK mouse event that fired this method call.
		"""
		selection = self.listEffects.get_selection().get_selected()
		
		# return if there is no active selection
		if not selection[1]:
			return
		
		# Create context menu on a right-click
		if mouse.button == 3:
			menu = gtk.Menu()
			
			menuItem = gtk.ImageMenuItem(_("_Activate effect"), True)
			menuItem.set_image(gtk.image_new_from_stock(gtk.STOCK_GO_FORWARD, gtk.ICON_SIZE_MENU))
			menuItem.show()
			menu.append(menuItem)
			menuItem.connect("activate", self.OnEffectAdd)
		
			menu.popup(None, None, None, mouse.button, mouse.time)
			
	#_____________________________________________________________________
	
	def OnEffectActivated(self, treeview, path, view_column):
		"""
		Adds the double clicked effect from the left effects pane to the Instrument.
		
		Parameters:
			treeview -- treeview that fired this event.
			path -- path to the activated row. Format: (index,)
			view_column -- the column in the activated row.
		"""
		self.OnEffectAdd()
		
	#_____________________________________________________________________
	
	def OnActiveEffectActivated(self, treeview, path, view_column):
		"""
		Displays the settings for the double clicked effect from the active effects
		pane.
		
		Parameters:
			treeview -- treeview that fired this event.
			path -- path to the activated row. Format: (index,)
			view_column -- the column in the activated row.
		"""
		self.OnEffectSettings()
	
	#_____________________________________________________________________
	
	def OnEffectAdd(self, widget=None):
		"""
		Adds the currently selected effect to the Instrument.
		
		Parameters:
			widget -- reserved for GTK callbacks, don't use it explicitly.
		"""
		selection = self.listEffects.get_selection().get_selected()
		
		# return if there is no active selection
		if not selection[1]:
			return

		self.instrument.AddEffect(self.filterEffects[selection[1]][1])
		
	#_____________________________________________________________________

	def OnEffectUp(self, button):
		"""
		Moves the selected effect one position up on the list.
		
		Parameters:
			button -- reserved for GTK callbacks, don't use it explicitly.
		"""
		selection = self.listActiveEffects.get_selection().get_selected()
		
		# return if there is no active selection or it's the first element
		if not selection[1] or self.modelActiveEffects[selection[1]].path == (0,):
			return
		
		# grab references to the effect position in the table and the
		# effect element itself, to then move it
		effectPos = self.modelActiveEffects[selection[1]].path[0]
		effect = self.instrument.effects[effectPos]
		
		newPosition = self.modelActiveEffects[selection[1]].path[0]-1
		self.instrument.ChangeEffectOrder(effect, newPosition)
		
		# reselect the effect moved in this operation
		self.listActiveEffects.set_cursor(newPosition)
	
	#_____________________________________________________________________
	
	def OnEffectDown(self, button):
		"""
		Moves the selected effect one position down on the list.
		
		Parameters:
			button -- reserved for GTK callbacks, don't use it explicitly.
		"""
		selection = self.listActiveEffects.get_selection().get_selected()
		
		# return if there is no active selection or it's the last element
		if not selection[1] or self.modelActiveEffects[selection[1]].next == None:
			return
		
		# grab references to the effect position in the table and the
		# effect element itself, to then move it
		effectPos = self.modelActiveEffects[selection[1]].path[0]
		effect = self.instrument.effects[effectPos]
		
		newPosition = self.modelActiveEffects[selection[1]].path[0]+1
		self.instrument.ChangeEffectOrder(effect, newPosition)
		
		# reselect the effect moved in this operation
		self.listActiveEffects.set_cursor(newPosition)
		
	#_____________________________________________________________________
	
	def OnEffectDeleted(self, widget):
		"""
		Removes the currently selected effect from the Instrument.
		
		Parameters:
			widget -- reserved for GTK callbacks, don't use it explicitly.
		"""
		selection = self.listActiveEffects.get_selection().get_selected()
		
		# return if there is no active selection
		if not selection[1]:
			return
		
		# grab references to the effect position in the table and the
		# effect element itself, to then delete it
		effectPos = self.modelActiveEffects[selection[1]].path[0]
		effect = self.instrument.effects[effectPos]
		
		self.instrument.RemoveEffect(effect)
		
		# select another existing active effect for usability,
		# so users can repeteadly click on delete to delete all effects
		if len(self.modelActiveEffects) > 0:
			self.listActiveEffects.set_cursor(0)
		
	#_____________________________________________________________________
	
	def OnEffectSettings(self, button=None):
		"""
		Show a dialog filled with settings sliders for a specific effect
		
		Parameters:
			button -- reserved for GTK callbacks, don't use it explicitly.
		"""
		# TODO: Make this modal or as part of the effects window
		selection = self.listActiveEffects.get_selection().get_selected()
		
		# return if there is no active selection
		if not selection[1]:
			return
		
		# grab references to the effect position in the table and the
		# effect element itself
		self.effectpos = self.modelActiveEffects[selection[1]].path[0]
		self.effectelement = self.instrument.effects[self.effectpos]
		
		# this variable is used to slash the values for the different sliders
		self.sliderdict = {}
		
		# Threshold value to determine if a property is buggy
		self.propertyThreshold = 1000000
		
		# Min/Max value assigned to buggy properties
		self.fixValue = 15.0
		
		# set the index of the current edited effect - used to reference the
		# effect elsewhere
		self.settWin = gtk.glade.XML(Globals.GLADE_PATH, "EffectSettingsDialog")

		settsignals = {
			"on_presetsCombo_changed" : self.OnSingleEffectPresetChanged,
			"on_savePresetButton_clicked" : self.OnSingleEffectPresetSave,
			"on_deletePresetButton_clicked" : self.OnSingleEffectPresetDelete,
			"on_closeButton_clicked" : self.OnSingleEffectSettingsClose
		}

		self.settWin.signal_autoconnect(settsignals)

		# create references to glade items
		self.settingswindow = self.settWin.get_widget("EffectSettingsDialog")
		self.settingsvbox = self.settWin.get_widget("EffectSettingsVBox")
		self.effectLabel = self.settWin.get_widget("effectLabel")
		self.effectImage = self.settWin.get_widget("effectImage")
		self.settingstable = self.settWin.get_widget("settingsTable")
		self.presetscombo = self.settWin.get_widget("presetsCombo")
		
		self.settingsHeaderCairoImage = CairoDialogHeaderImage(_("Effect Settings"))
		self.settingsHeaderCairoImage.set_size_request(450, 60)
		self.settingsvbox.pack_start(self.settingsHeaderCairoImage, expand=False, fill=True)
		self.settingsHeaderCairoImage.show()
		
		# set the window icon and parent (for correct modal mode)
		self.settingswindow.set_icon(self.windowIcon)
		self.settingswindow.set_transient_for(self.window)
		
		# tooltips object used to assign tooltips to the sliders
		tooltips = gtk.Tooltips()
		
		# grab a list of properties from the effect
		proplist = gobject.list_properties(self.instrument.effects[self.effectpos])
		
		# set the effect name and icon
		self.effectLabel.set_label("<b>%s</b>" % self.modelActiveEffects[selection[1]][2])
		self.effectImage.set_from_pixbuf(self.modelActiveEffects[selection[1]][0])
		
		# resize the settingstable to accomodate the number of settings sliders required
		self.settingstable.resize(len(proplist), 2)
		
		count = 0
		
		# iterate through the properties list, determine its value type 
		# and show it where needed
		for property in proplist:
			# non readable params
			if not(property.flags & gobject.PARAM_READABLE):
				# create and attach a property name label to the settings table
				label = gtk.Label(property.name)
				label.set_alignment(1, 0.5)
				self.settingstable.attach(label, 0, 1, count, count+1, xoptions=gtk.FILL)

				rlabel = gtk.Label(_("-parameter not readable-"))
				self.settingstable.attach(rlabel, 1, 2, count, count+1)

			# just display non-writable param values
			elif not(property.flags & gobject.PARAM_WRITABLE):
				# create and attach a property name label to the settings table
				label = gtk.Label(property.name)
				label.set_alignment(1, 0.5)
				self.settingstable.attach(label, 0, 1, count, count+1, xoptions=gtk.FILL)
								
				wvalue = self.effectelement.get_property(property.name)
				
				if wvalue != None:
					wlabel = gtk.Label(wvalue)
					self.settingstable.attach(wlabel, 1, 2, count, count+1, xoptions=gtk.FILL)

			elif hasattr(property, "minimum") and hasattr(property, "maximum"):
				# create and attach a property name label to the settings table
				label = gtk.Label(property.name)
				label.set_alignment(1, 0.5)
				self.settingstable.attach(label, 0, 1, count, count+1, xoptions=gtk.FILL)

				# assume that it's numeric so we can use an HScale
				value = self.effectelement.get_property(property.name)
				
				# use these values for the slider range
				minValue = property.minimum
				maxValue = property.maximum
				
				# fix the ridiculously big min/max range values if needed
				if property.minimum < -self.propertyThreshold:
					minValue = -self.fixValue
					
				if property.maximum > self.propertyThreshold:
					maxValue = self.fixValue
				
				adj = gtk.Adjustment(value, minValue, maxValue)
				adj.connect("value_changed", self.SetSingleEffectSetting, property.name, property)
				self.sliderdict[property.name] = hscale = gtk.HScale(adj)
				hscale.set_value_pos(gtk.POS_RIGHT)
				
				# add step increment for mouse-wheel scrolling - proportional
				# to range in view. Must be at least the smallest interval
				# possible (determined by get_digits()) or scroll doesn't happen
				adj.step_increment = (maxValue - minValue) / 100.0
				adj.step_increment = max(adj.step_increment, 1.0 / (10 ** hscale.get_digits()))
				
				# check for ints and change digits
				if not((property.value_type == gobject.TYPE_FLOAT) or
					   (property.value_type == gobject.TYPE_DOUBLE)):
					hscale.set_digits(0)

				# add the slider to the settings table (with tooltips)
				tooltips.set_tip(self.sliderdict[property.name], property.blurb)
				self.settingstable.attach(self.sliderdict[property.name], 1, 2, count, count+1, gtk.FILL|gtk.EXPAND)
			
			count += 1

		# set up presets
		self.elementfactory = self.effectelement.get_factory().get_name()
		
		self.model = gtk.ListStore(str)
		self.presetscombo.set_model(self.model)
		self._LoadEffectPresets()
		
		# show the settings window	
		self.settingstable.show()
		self.settingswindow.show_all()

	#_____________________________________________________________________
	
	def _LoadEffectPresets(self):
		"""
		Loads the presets for the selected effect if there are any available.
		It then adds them to the effects presets combo box.
		
		The following checks are performed during this operation:
			a) the preset is on the system (in LADSPA_FACTORY_REGISTRY).
			b) the preset is for this plugin.
		"""
		self.availEffectPresets = []
		self.model.clear()
		
		self.presets.FillEffectsPresetsRegistry()
		
		try:
			effectPresets = self.presets.effectpresetregistry["effects"][self.elementfactory]
			if self.elementfactory in Globals.LADSPA_FACTORY_REGISTRY:
				for pres in effectPresets:
					deps = effectPresets[pres]["dependencies"]
					if len(deps) == 1 and self.elementfactory in deps:
						self.availEffectPresets.append(pres)
						self.model.append([pres])
						#self.presetscombo.append_text(pres)
		except KeyError:
			# no presets for this effect
			pass
	
	#_____________________________________________________________________
		
	def OnSingleEffectSettingsClose(self, button):
		"""
		Close the effect settings window.
		
		Parameters:
			button -- reserved for GTK callbacks, don't use it explicitly.
		"""
		self.settingswindow.destroy()

	#_____________________________________________________________________	
	
	def SetSingleEffectSetting(self, slider, name, property):
		"""
		Set the value of a gstreamer property from an effects slider.
		
		Parameters:
			slider -- slider control which indicates the value to assign to property.
			name -- name of the property being set.
			property -- property being set.
		"""
		# grab the slider setting
		if not((property.value_type == gobject.TYPE_FLOAT) or
			   (property.value_type == gobject.TYPE_DOUBLE)):
			value = int(slider.get_value())
		else:
			value = slider.get_value()
		
		# set the gstreamer property
		self.instrument.effects[self.effectpos].set_property(name, value)
		
	#_____________________________________________________________________	

	def OnSingleEffectPresetSave(self, widget):
		"""
		Grab the effect properties and send it to the presets code
		to be saved.
		
		Parameters:
			widget -- reserved for GTK callbacks, don't use it explicitly.
		"""
		# grab the label from the combo
		label = self.presetscombo.get_active_text()
		
		if label == "":
			return
		
		effectdict = {}
		
		effect = self.instrument.effects[self.effectpos]
		effectelement = effect.get_factory().get_name()
		
		proplist = gobject.list_properties(effect)
		
		# for each property in proplist, add it to the effectdict dictionary
		for property in proplist:
			effectdict[property.name] = effect.get_property(property.name)

		# save the preset	
		self.presets.SaveSingleEffect(label, effectdict, effectelement, "LADSPA")
		
	#_____________________________________________________________________
	
	def OnSingleEffectPresetDelete(self, button):
		"""
		Removes the selected effect preset.
		
		Parameters:
			button -- reserved for GTK callbacks, don't use it explicitly.
		"""
		presetName = self.presetscombo.get_active_text()
		
		if presetName == "":
			return
		
		effect = self.instrument.effects[self.effectpos]
		effectName = effect.get_factory().get_name()
		
		self.presets.DeleteSingleEffect(presetName, effectName)
	
	#_____________________________________________________________________
	
	def OnEffectChainPresetSave(self, button):
		"""
		Grabs the effects chain and sends it to the presets code to be saved.
		
		Parameters:
			button -- reserved for GTK callbacks, don't use it explicitly.
		"""
		# grab the label from the combo
		label = self.comboPresets.get_active_text()
		
		if label == "":
			return
		
		self.effectlist = []
				
		for effect in self.instrument.effects:
			effectdict = {}
			effectsettings = {}

			proplist = gobject.list_properties(effect)
		
			for property in proplist:
				effectsettings[property.name] = effect.get_property(property.name)

			effectdict["effectelement"] = effect.get_factory().get_name()
			effectdict["effecttype"] = "LADSPA"
			effectdict["settings"] = effectsettings
			
			self.effectlist.append(effectdict)
		
		self.presets.SaveEffectChain(label, self.effectlist, self.instrument.instrType)

	#_____________________________________________________________________
	
	def OnEffectChainPresetDelete(self, button):
		"""
		Removes the selected effects chain preset.
		
		Parameters:
			button -- reserved for GTK callbacks, don't use it explicitly.
		"""
		presetName = self.comboPresets.get_active_text()
		
		if presetName == "":
			return
		
		self.presets.DeleteEffectChain(presetName, self.instrument.instrType)
	
	#_____________________________________________________________________
	
	def OnSingleEffectPresetChanged(self, combo):
		"""
		Loads a preset when it's selected from the single effect preset combo.
		
		Parameters:
			combo -- reserved for GTK callbacks, don't use it explicitly.
		"""
		presetname = name = combo.get_active_text()
		
		if presetname not in self.availEffectPresets:
			return
		
		settings = self.presets.LoadSingleEffect(presetname, 
												 self.effectelement.get_factory().get_name())
		if not settings:
			return

		for item in settings:
			#self.sliderdict[str(item)]
			try:
				self.effectelement.set_property(item, settings[item])
				self.sliderdict[str(item)].set_value(settings[item])
			except:
				pass
		
		self.settingstable.show()
		self.settingswindow.show_all()

	#_____________________________________________________________________	
	
	def OnEffectChainPresetChanged(self, combo):
		"""
		Loads a preset when it's selected from the chain preset combo.
		
		Parameters:
			combo -- reserved for GTK callbacks, don't use it explicitly.
		"""
		# If we're still setting up the gui then we dont want to *do* anything when 
		# we set the combobox so check the flag and quit if necessary
		if self.updatinggui == True:
			return
		
		presetname = name = combo.get_active_text()
		if presetname not in self.availpresets:
			return

		self.instrument.currentchainpreset = combo.get_active()
		settings = self.presets.LoadEffectChain(presetname, self.instrument.instrType)
		
		self.instrument.effects = []
		
		for item in settings:
			effect = settings[item]
			prefs = effect['preferences']
			setts = effect['settings']

			effect = self.instrument.AddEffect(prefs['effectelement'])

			for sett in setts:
				try:
					setts[sett]
					effect.set_property(sett, setts[sett])

				except:
					Globals.debug("problem loading preset property %s= %s"%(sett,setts[sett]))
					pass

		self.Update()
		
	#_____________________________________________________________________	
	
	def BringWindowToFront(self):
		"""
		Puts the InstrumentsEffectDialog on top of other windows.
		"""
		self.window.present()
		
	#_____________________________________________________________________



	def OnResize(self, widget, event):
		"""
		This method is called when the Instrument Effect dialog is resized

		Parameters:
			widget -- GTK callback parameter.
			event -- GTK callback parameter.
			
		Returns:
			False -- continue GTK signal propagation.
		"""	

		(self.width, self.height) = widget.get_size()

		return False

	#_____________________________________________________________________


	def OnDestroy(self, widget=None, event=None):
		"""
		Called when the instrument effects dialog is destroyed

		Parameters: 
			widget -- reserved for GTK callbacks, don't use it explicitly.
			event -- reserved for GTK callbacks, don't use it explicitly.
		"""
		
		Globals.settings.general["instrumenteffectwindowwidth"] = self.width
		Globals.settings.general["instrumenteffectwindowheight"] = self.height
		Globals.settings.write()
	#_____________________________________________________________________	

#=========================================================================

class CairoDialogHeaderImage(gtk.DrawingArea):
	"""
	Renders a banner for the top of the effects dialog using Cairo.
	"""
	
	_LEFT_GRADIENT_STOP_RGBA = (0, 0.99, 0.87, 0.38, 1)
	_RIGHT_GRADIENT_STOP_RGBA = (1, 0.93, 0.55, 0.16, 1)
	_TEXT_RGBA = (0.93, 0.55, 0.16, 1)
	_FONT_SIZE = 40
	_LEFT_LOGO_INDENT = 7
	_LEFT_TEXT_INDENT = 56
	
	#_____________________________________________________________________
	
	def __init__(self, headerText):
		"""
		Creates a new instance of CairoDialogHeaderImage.
		
		Parameters:
			headerText -- text to be displayed at the header.
		"""
		gtk.DrawingArea.__init__(self)
		self.headerText = headerText
		
		self.connect("configure_event", self.OnSizeChanged)
		self.connect("expose-event", self.OnDraw)
		
		self.logo = cairo.ImageSurface.create_from_png(os.path.join(Globals.IMAGE_PATH, "jokosher.png"))
		self.source = cairo.ImageSurface(cairo.FORMAT_ARGB32, self.allocation.width, self.allocation.height)
		self.GenerateCachedImage()
		
	#_____________________________________________________________________
		
	def OnSizeChanged(self, obj, evt):
		"""
		Called when the widget's size changes.
		
		Parameters:
			obj -- reserved for Cairo callbacks, don't use it explicitly. *CHECK*
			evt -- reserved for Cairo callbacks, don't use it explicitly. *CHECK*
		"""
		if self.allocation.width != self.source.get_width() or self.allocation.height != self.source.get_height():
			self.source = cairo.ImageSurface(cairo.FORMAT_ARGB32, self.allocation.width, self.allocation.height)
			self.GenerateCachedImage()

	#_____________________________________________________________________

	def GenerateCachedImage(self):
		"""
		Renders the image so that we don't have to re-render it every
		time there is an expose event from a mouse move or something.
		"""
		
		rect = self.get_allocation()

		ctx = cairo.Context(self.source)
		ctx.set_antialias(cairo.ANTIALIAS_SUBPIXEL)
		
		# Create our gradient
		pat = cairo.LinearGradient(0.0, 0.0, rect.width, 0)
		pat.add_color_stop_rgba(*self._RIGHT_GRADIENT_STOP_RGBA)
		pat.add_color_stop_rgba(*self._LEFT_GRADIENT_STOP_RGBA)

		# Fill the widget
		ctx.rectangle(0, 0, rect.width, rect.height)
		ctx.set_source(pat)
		ctx.fill()
		
		#Paint the Jokosher logo on top (x=7, y=8 centres the logo)
		logoPos = (rect.height - self.logo.get_height()) / 2
		ctx.set_source_surface(self.logo, self._LEFT_LOGO_INDENT, int(logoPos))
		ctx.paint()
		
		ctx.set_source_rgba(*self._TEXT_RGBA)
		textPos = rect.height - ((rect.height - self._FONT_SIZE) / 2) - 5
		ctx.move_to(self._LEFT_TEXT_INDENT, int(textPos))
		ctx.set_font_size(self._FONT_SIZE)
		ctx.show_text(self.headerText)
		ctx.stroke()
	
	#_____________________________________________________________________

	def OnDraw(self, widget, event):
		"""
		Handles the GTK expose event.
		
		Parameters:
			widget -- reserved for GTK callbacks, don't use it explicitly.
			event -- reserved for GTK callbacks, don't use it explicitly.
			
		Returns:
			False -- indicates the GTK signal to:
					1) continue propagating the regular signal.
					2) stop calling the callback on a timeout_add.
		"""
		
		ctx = widget.window.cairo_create()
		#Don't repaint the whole area unless necessary
		ctx.rectangle(event.area.x, event.area.y, event.area.width, event.area.height)
		ctx.clip()
		# Paint the cached image across the widget
		ctx.set_source_surface(self.source, 0, 0)	
		ctx.paint()

		return False
		
	#_____________________________________________________________________


#=========================================================================



#
#	THIS FILE IS PART OF THE JOKOSHER PROJECT AND LICENSED UNDER THE GPL. SEE
#	THE 'COPYING' FILE FOR DETAILS
#
#	CompactMixView.py
#	
#	A sub-class of gtk.Frame containing the mixing view of the project
#
#-------------------------------------------------------------------------------

import gtk
import gobject
import RecordingView
from MixerStrip import *
from MasterMixerStrip import *
import gettext
_ = gettext.gettext

#=========================================================================

class CompactMixView(gtk.Frame):
	"""
	This class implements the mixing workspace view.
	"""
	
	""" Number of times a second the VUWidgets need updating. """
	FPS = 10
	
	#_____________________________________________________________________
	
	def __init__(self, project, mainview):
		"""
		Creates a new instance of CompactMixView.
		
		Parameters:
			project -- the active Project.
			mainview -- reference to the MainApp Jokosher window.
		"""
		gtk.Frame.__init__(self)
		self.project = project
		self.mainview = mainview
		self.small = True
		self.mix = False
		self.mixerStripList = []
		self.minimisedButtonList = []
		self.lanes = []
		self.Updating = False
		self.CreateInstrumentBar()
		
		self.vbox = gtk.VBox()
		self.add(self.vbox)
		self.hbox = gtk.HBox()
		self.vbox.pack_start(self.hbox, True, True)
		
		self.mastermixer = MasterMixerStrip(self.project, self, self.mainview)
		self.hbox.pack_end(self.mastermixer, False, False)
		
		self.scrollWindow = gtk.ScrolledWindow()
		self.mixerBox = gtk.HBox()
		
		self.scrollWindow.add_with_viewport(self.mixerBox)
		# remove the shadow on the viewport
		self.scrollWindow.child.set_shadow_type(gtk.SHADOW_NONE)
		self.scrollWindow.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_NEVER)
		self.hbox.pack_start(self.scrollWindow, True, True)
		

		self.project.connect("instrument::added", self.OnInstrumentAdded)
		self.project.connect("instrument::reordered", self.OnInstrumentReordered)
		self.project.connect("instrument::removed", self.OnInstrumentRemoved)
		
		#initialize the instrument widgets
		for instr in self.project.instruments:
			self.OnInstrumentAdded(self.project, instr)
		self.show_all()
		self.UpdateTimeout = False
		
	#_____________________________________________________________________

	def CreateInstrumentBar(self):
		self.instrumentBar = gtk.Toolbar()
		self.instrumentBar.set_show_arrow(True)
		self.instrumentBar.set_style(gtk.TOOLBAR_BOTH_HORIZ)
		toollab = gtk.ToolItem()
		lab = gtk.Label()
		lab.set_markup(_("<b>Instruments Not Shown:</b>"))
		toollab.add(lab)
		toollab.set_is_important(True)
		self.instrumentBar.insert(toollab, 0)
	
	#_____________________________________________________________________

	def OnMinimiseTrack(self, widget, instr):
		"""
		Minimizes a mixer strip (instrument).
		
		Parameters:
			widget -- reserved for GTK callbacks, don't use it explicitly.
			instr -- the Instrument to be hidden.
		"""
		instr.SetVisible(False)
		
	#_____________________________________________________________________

	def OnMaximiseTrack(self, widget, instr):
		"""
		Maximizes a mixer strip (instrument).
		
		Parameters:
			widget -- reserved for GTK callbacks, don't use it explicitly.
			instr -- the Instrument to be shown.
		"""
		instr.SetVisible(True)
	
	#_____________________________________________________________________
	
	def OnInstrumentVisible(self, instrument):
		"""
		Callback for when the visible status of an instrument changes.
		
		Parameters:
			instrument -- the instrument instance that send the signal.
		"""
		
		visib = instrument.isVisible
		for strip in self.mixerStripList:
			if not strip.instrument is instrument:
				continue
			
			if visib and not strip.parent:
				self.mixerBox.pack_start(strip, False, False)
				strip.show_all()
			elif not visib and strip.parent:
				self.mixerBox.remove(strip)
			
			break
		
		for instr, toolButton in self.minimisedButtonList:
			if not instr is instrument:
				continue
				
			if not visib and not toolButton.parent:
				self.instrumentBar.insert(toolButton, -1)
				toolButton.show_all()
			elif visib and toolButton.parent:
				self.instrumentBar.remove(toolButton)
			
			break
		
		# Only show the instrument bar if there is something minimized
		minimisedInstrs = [x for x in self.project.instruments if not x.isVisible]
		if minimisedInstrs and not self.instrumentBar.parent:
			self.vbox.pack_end(self.instrumentBar, False, True)
			self.instrumentBar.show_all()
		elif not minimisedInstrs and self.instrumentBar.parent:
			self.vbox.remove(self.instrumentBar)
		
	#_____________________________________________________________________
	
	def OnInstrumentAdded(self, project, instrument):
		"""
		Callback for when an instrument is added to the project.
		
		Parameters:
			project -- The project that the instrument was added to.
			instrument -- The instrument that was added.
		"""
		
		strip = MixerStrip(self.project, instrument, self, self.mainview)
		strip.connect("minimise", self.OnMinimiseTrack, instrument)
		self.mixerStripList.append(strip)
		
		#create the toolbar button that will be shown when the instrument is minimised
		imgsize = gtk.icon_size_lookup(gtk.ICON_SIZE_MENU)[0]
		pixbuf = instrument.pixbuf.scale_simple(imgsize, imgsize, gtk.gdk.INTERP_BILINEAR)
		image = gtk.Image()
		image.set_from_pixbuf(pixbuf)
		toolButton = gtk.ToolButton()
		toolButton.set_label(instrument.name)
		toolButton.set_icon_widget(image)
		toolButton.set_is_important(True)
		toolButton.connect("clicked", self.OnMaximiseTrack, instrument)
		
		self.minimisedButtonList.append( (instrument, toolButton) )
		
		instrument.connect("visible", self.OnInstrumentVisible)
		#check if the instrument is currently visible and show the widgets
		self.OnInstrumentVisible(instrument)
	
	#_____________________________________________________________________
	
	def OnInstrumentRemoved(self, project, instrument):
		"""
		Callback for when an instrument is removed from the project.
		
		Parameters:
			project -- The project that the instrument was removed from.
			instrument -- The instrument that was removed.
		"""
		
		for strip in self.mixerStripList:
			if strip.instrument is instrument:
				if strip.parent:
					self.mixerBox.remove(strip)
				strip.Destroy()
				self.mixerStripList.remove(strip)
				break
				
		for instr, toolButton in self.minimisedButtonList:
			if instr is instrument:
				if toolButton.parent:
					self.instrumentBar.remove(toolButton)
				self.minimisedButtonList.remove( (instr, toolButton) )
				break
	
	#_____________________________________________________________________
	
	def OnInstrumentReordered(self, project, instrument):
		"""
		Callback for when an instrument's position in the project has changed.
		
		Parameters:
			project -- The project that the instrument was changed on.
			instrument -- The instrument that was reordered.
		"""
		if not instrument.isVisible:
			return
		
		visibleInstrs = [x for x in self.project.instruments if x.isVisible]
		for strip in self.mixerStripList:
			if strip.instrument is instrument:
				if strip.parent:
					pos = visibleInstrs.index(instrument)
					self.mixerBox.reorder_child(strip, pos + 1)
				break
		
	#_____________________________________________________________________
	
	def OnUpdateTimeout(self):
		"""
		Called at intervals (self.FPS) to update the VU meters.
		
		Returns:
			True -- keeps the timeout going during playback.
			False -- stops the timeout when playback stops.
		"""
		if self.mainview.isPlaying:
			self.mastermixer.vu.queue_draw()
			
			# redraw VU widgets for each instrument
			for mix in self.mixerStripList:
				mix.vu.queue_draw()
			
			return True
		else:
			# kill timeout when playback has stopped
			self.UpdateTimeout = False
			return False
	
	#_____________________________________________________________________
	
	def StartUpdateTimeout(self):
		""" 
		Initiates the OnUpdateTimeout - called from MainApp.play()
		when the play button is pressed.
		"""
		if not self.UpdateTimeout:
			gobject.timeout_add(int(1000 / self.FPS), self.OnUpdateTimeout, priority = gobject.PRIORITY_LOW)
			self.UpdateTimeout = True
	
#=========================================================================
		



#
#	THIS FILE IS PART OF THE JOKOSHER PROJECT AND LICENSED UNDER THE GPL. SEE
#	THE 'COPYING' FILE FOR DETAILS
#
#	ExtensionManager.py
#	
#	This module defines the ExtensionManager class which is 
#	responsible for controlling extensions to Jokosher
#
#-------------------------------------------------------------------------------
import Globals
import Extension
import pkg_resources
import os.path
import imp
import gtk
import shutil

#=========================================================================

class ExtensionManager:
	"""
	The ExtensionManager class handles the installation and running of
	Extensions. It also controls their disabling and removal.
	"""
	#_____________________________________________________________________
	
	def __init__(self, mainapp):
		"""
		Creates a new Instance of ExtensionManager.
		
		Parameters:
			parent -- the parent MainApp Jokosher window.
		"""
		self.mainapp = mainapp
		self.loadedExtensions = []
		self.API = Extension.ExtensionAPI(mainapp)
		self.LoadAllExtensions()
		
	#_____________________________________________________________________
	
	def register(self, extension, filename, directory, local):
		"""
		Called from Extension.LoadAllExtensions afer the Extensions
		module has been imported (and the class instantiated in the case
		of extensions that are eggs). 
		
		Parameters:	
			extension -- a reference to the Extension. Either a module 
						or in the case of an Extension imported from an
						egg, an instance object.
			filename -- the name of the file containing the Extension.
			directory -- full path to the directory containing the file.
			local --	True = the extension is to be copied to the
								local extension directory after checking
						False = don't copy the Extension.
		
		Returns:
			True -- the Extension was successfully registered.
			False -- an error ocurred while trying to register the Extension,
					or the Extension has been disabled via the ExtensionManagerDialog.
		"""
		name = None
		preferences = False
		# check if the necessary attributes EXTENSION_NAME, EXTENSION_VERSION
		#	and EXTENSION_DESCRIPTION are present and
		#	refuse to start the extension if they are missing. 
		requiredAttributes = ("EXTENSION_NAME", "EXTENSION_VERSION",
				"EXTENSION_DESCRIPTION", "startup", "shutdown")
		missingAttrs = []
		for attr in requiredAttributes:
			if not hasattr(extension, attr):
				missingAttrs.append(attr)
		
		if missingAttrs:
			Globals.debug("\t" + filename, "missing", ", ".join(missingAttrs))
			return False
		else:
			name = extension.EXTENSION_NAME
			version = extension.EXTENSION_VERSION
			description = extension.EXTENSION_DESCRIPTION

		# check for preferences attribute
		if hasattr(extension, "preferences"):
			preferences = True

		# check extension is not already loaded
		for testExtension in self.loadedExtensions:
			if testExtension["name"] == name:
				Globals.debug(filename + " extension '" + name + "' already present")
				return False
			
		# find full file name of extension
		extensionFile = os.path.join(directory, filename)
		
		# if we are installing locally first check the file doesn't exist and
		# then try and copy it 
		if local:
			extensionFile = os.path.join(Globals.JOKOSHER_DATA_HOME, "extensions", filename)
			if os.path.exists(extensionFile):
				Globals.debug("Filename " + extensionFile + " already exists")
				return False
			else:
				try:
					shutil.copy(os.path.join(directory, filename), extensionFile)
				except Exception, e:
					Globals.debug(filename + "Failed copying file: " + str(e))
					return False
		
		# check if extension's requirements are met (only if the extension requires it)
		testResults = (True, "")
		try:
			if hasattr(extension, "check_dependencies"):
				testResults = extension.check_dependencies()
		except Exception, e:
			Globals.debug(name + " extension could not check its dependencies")
			Globals.debug(e)
			return False
		
		# if the system doesn't provide what the extension needs,
		# fail loading this plugin and set the error message
		if testResults[0] == False:
			#TODO: inform the user of the error
			Globals.debug(name + ": "+testResults[1])
			return False
		
		# check if the extension is blacklisted, if so, mark it as disabled
		enabled = True
		if name in Globals.settings.extensions['extensions_blacklist']:
			enabled = False

		# add details to loadedExtensions list
		self.loadedExtensions.append(
			{"name":name, 
			 "description":description,
			 "version":version,
			 "extension":extension,
			 "enabled":enabled,
			 "preferences":preferences,
			 "filename":extensionFile })
			 
		return True
		
	#_____________________________________________________________________
	
	def GetExtensions(self):
		"""
		Obtain a generator for iterating the list of loadedExtensions.
		
		Returns:
			a generator with the list of loadedExtensions.
		"""
		return iter(self.loadedExtensions)

	#_____________________________________________________________________
	
	def LoadExtensionFromFile(self, filename, directory, local=False):
		"""
		Tries to load an Extension fron a given file.
		
		Parameters:
			filename -- the name of the file containing the Extension.
			directory -- full path to the directory containing the file.
			local --	True = the extension is to be copied to the
								local extension directory after checking
						False = don't copy the Extension.
		
		Returns:
			True -- the Extension was successfully loaded.
			False -- an error ocurred while trying to load the Extension,
					or the Extension has been disabled via the ExtensionManagerDialog.
		"""
		Globals.debug("\timporting extension", filename)
		extension = None
		if filename.endswith(".py"):
			# for a python module try and import it
			extension = None
			fn = os.path.splitext(filename)[0]
			exten_file, filename, description = imp.find_module(fn, [directory])
			
			try:
				extension = imp.load_module(fn, exten_file, filename, description)
			except Exception, e:
				Globals.debug("\t\t...failed.")
				Globals.debug(e)
				if exten_file:
					exten_file.close()
				return False
			if exten_file:
				exten_file.close()
				
		elif filename.endswith(".egg"):
			# for an egg, add it to working_set and then try and
			# load it and pick out the entry points
			fullName = os.path.join(directory, filename)
			pkg_resources.working_set.add_entry(fullName)
			dist_generator = pkg_resources.find_distributions(fullName)
			for dist in dist_generator:
				try:
					extension_class = dist.load_entry_point("jokosher.extensions", "extension")
					# create an instance of the class
					extension = extension_class()
				except Exception, e :
					Globals.debug("\t\t...failed.")
					Globals.debug(e)
					return False
		else:
			# any other file extension is wrong
			Globals.debug("Invalid extension file suffix for", filename)
			return False
		
		# try and register the extension - quit if failed
		if not self.register(extension, filename, directory, local):
			return False
		
		# if we're still here then start the extension, if its not in the extensions_blacklist
		if extension.EXTENSION_NAME not in Globals.settings.extensions['extensions_blacklist']:
			dir
			if not self.StartExtension(self.loadedExtensions[len(self.loadedExtensions)-1]["filename"]):
				return False
		
		return True
		
	#_____________________________________________________________________

	def StopExtension(self, filename):
		"""
		Stops the given Extension.
		
		Considerations:
			This method executes the shutdown() function of the Extension.
			This is mainly for disabling Extensions	on the fly, but is also
			used for removing them.
			
		Parameters:
			filename -- the name of the file containing the Extension.
			
		Returns:
			True -- the Extension was successfully stopped.
			False -- an error ocurred while trying to stop the Extension.
		"""
		for extension in self.GetExtensions():
			if extension['filename'] == filename:
				try:
					extension['extension'].shutdown()
				except Exception, e:
					Globals.debug(filename + " extension failed to shut down")
					Globals.debug(e)
					return False
		return True

	#_____________________________________________________________________

	def StartExtension(self, filename):
		"""
		Starts the given Extension.
		
		Considerations:
			This method executes the startup() function of the Extension
			This is mainly for enabling an Extension on the fly without
			loading another instance.
			
		Parameters:
			filename -- the name of the file containing the Extension.
			
		Returns:
			True -- the Extension was successfully started.
			False -- an error ocurred while trying to start the Extension.
		"""
		for extension in self.GetExtensions():
			if extension['filename'] == filename:
				try:
					extension['extension'].startup(self.API)
				except Exception, e:
					Globals.debug(filename + " extension failed to start")
					Globals.debug(e)
					return False
		return True

	#_____________________________________________________________________
	
	def RemoveExtension(self, filename):
		"""
		Removes the given Extension.
		
		Considerations:
			This function "unloads" the Extension. It executes the shutdown()
			function of the Extension and then removes it from loadedExtensions.
			
		Parameters:
			filename -- the name of the file containing the Extension.
			
		Returns:
			True -- the Extension was successfully removed.
			False -- an error ocurred while trying to remove the Extension.
		"""
		self.StopExtension(filename)
		index = -1
		for extension in self.GetExtensions():
			index += 1
			if extension['filename'] == filename:
				try:
					os.remove(filename)
					self.loadedExtensions.pop(index)
				except Exception, e:
					Globals.debug("Failed to remove " + filename)
					Globals.debug(e)
					return False
		return True
	
	#_____________________________________________________________________
	
	def ExtensionPreferences(self, filename):
		"""
		Loads the preferences() function of an Extension.
		
		Parameters:
			filename -- the name of the file containing the Extension.
			
		Returns:
			True -- the Extension's preferences were successfully loaded.
			False -- an error ocurred while trying to load the Extension's
					preferences.
		"""
		for extension in self.GetExtensions():
			if extension['filename'] == filename:
				try:
					extension['extension'].preferences()
				except:
					Globals.debug("Someone screwed up their preferences function")
					return False
		return True
				
	#_____________________________________________________________________
			
	def LoadAllExtensions(self):
		"""
		Load all the Extensions found in EXTENSION_PATHS and import every .py
		and .egg file found.
		"""
		Globals.debug("Loading extensions:")
		for exten_dir in Globals.EXTENSION_PATHS:
			if not os.path.isdir(exten_dir):
				continue
			for filename in os.listdir(exten_dir):
				if filename.endswith(".egg") or filename.endswith(".py"):
					self.LoadExtensionFromFile(filename, exten_dir)
					# don't block the gui when loading many extensions
					while gtk.events_pending():
						gtk.main_iteration()
		
	#_____________________________________________________________________
		
#=========================================================================




import gtk
import gobject
import cairo

#=========================================================================

class EventLaneHSeparator(gtk.HSeparator):
	def __init__(self, project, transport):
		gtk.HSeparator.__init__(self)
		
		self.x_pos = 0
		self.prev_x_pos = 0
		transport.connect("position", self.OnTransportPosition)
		project.connect("view-start", self.OnProjectViewStart)
	
	def OnProjectViewStart(self, project):
		self.OnTransportPosition(project.transport)
	
	def OnTransportPosition(self, transportManager, extraString=None):
		self.x_pos = transportManager.GetPixelPosition()
		
		a = self.get_allocation()
		
		self.queue_draw_area(a.x + self.prev_x_pos, a.y, 1, a.height)
		self.queue_draw_area(a.x + self.x_pos     , a.y, 1, a.height)
		
		self.prev_x_pos = self.x_pos
		
	def do_expose_event(self, event):
		retval = gtk.HSeparator.do_expose_event(self, event)
		
		context = self.window.cairo_create()
		
		context.set_antialias(cairo.ANTIALIAS_NONE)
		context.set_line_width(1)
		context.set_source_rgb(1, 0, 0)
		
		a = self.get_allocation()
		x = self.x_pos + a.x
		
		context.move_to(x + 0.5, a.y - 0.5)
		context.line_to(x + 0.5, a.y + a.height + 0.5)
		context.stroke()
		
		return retval
	
#=========================================================================

gobject.type_register(EventLaneHSeparator)



#
#	THIS FILE IS PART OF THE JOKOSHER PROJECT AND LICENSED UNDER THE GPL. SEE
#	THE 'COPYING' FILE FOR DETAILS
#
#	UndoSystem.py
#
#	Contains the decorator needed to allow other classes to hook specific
#	function calls into the undo stack.
#
#=========================================================================

def UndoCommand(*command, **command_options):
	"""
	Decorates functions, enabling them to be logged in the undo stack.
	The decorating process is transparent to the clients.
	
	Parameters:
		command -- the undo command list of strings.
		command_options -- key-value parameters to change options.
		
	Returns:
		an UndoFunction which decorates the original function.
	"""
	
	def UndoFunction(func):
		"""
		This is the actual decorator function. When decorated,
		this function will be called with func as the function to
		be decorated.
		
		Parameters:
			func -- the function to be decorated.
			
		Returns:
			an UndoWrapper to replace the function, so that when it
			is called, UndoWrapper will be called instead, and will:
				1)log the function call to the undo stack, and 
				2)call the function originally wanted.
		"""
		
		def UndoWrapper(funcSelf, *args, **kwargs):
			"""
			This function will wrap and take the place of the function
			that is being decorated. All arguments to the original function
			will be saved, and sent to the decorated function call.
			The funcSelf value must be the first parameter, because
			the first parameter will always be self, and it carries a
			reference to the decorated function's class.
			
			Considerations:
				All decorated undo functions *must* be in a class or this will fail.
				
			Parameters:
				funcSelf -- reference to the decorated function's class.
				*args -- parameters meant for the decorated function.
				**kwargs -- dictionary of keyword:value parameters
						containing the optional _undoAction_ parameter.
				_undoAction_ -- has to be passed as a key:value pair inside kwargs.
						The AtomicUndoAction object to append the 
						command to or None to create a default 
						AtomicUndoAction with only the one command.
			
			Returns:
				the wrapped function resulting value.
			"""
			if kwargs.has_key("_undoAction_"):
				atomicUndoObject = kwargs["_undoAction_"]
				del kwargs["_undoAction_"]
			else:
				atomicUndoObject = None
			
			do_incremental_save = True
			if command_options.has_key("incremental_save"):
				do_incremental_save = command_options["incremental_save"]
			
			
			try:
				result = func(funcSelf, *args, **kwargs)
			except CancelUndoCommand, e:
				return e.result
			
			project = None
			if isinstance(funcSelf, Project.Project):
				project = funcSelf
				objectString = "P"
			elif isinstance(funcSelf, Instrument.Instrument):
				project = funcSelf.project
				objectString = "I%d" % funcSelf.id
			elif isinstance(funcSelf, Event.Event):
				project = funcSelf.instrument.project
				objectString = "E%d" % funcSelf.id
			
			if do_incremental_save:
				inc = IncrementalSave.Action(objectString, func.__name__, args, kwargs)
				project.SaveIncrementalAction(inc)
				# testing: make sure loading produces an identical result
				assert inc.StoreToString() == IncrementalSave.Action.LoadFromString(inc.StoreToString()).StoreToString()
			
			if not atomicUndoObject and project:
				atomicUndoObject = project.NewAtomicUndoAction()
				
			if atomicUndoObject:
				paramList = []
				for param in command[1:]:
					try:
						value = getattr(funcSelf, param)
					except:
						continue
					else:
						paramList.append(value)
				
				atomicUndoObject.AddUndoCommand(objectString, command[0], paramList)
			
			return result
		
			#_____________________________________________________________________
		
		UndoWrapper.wrapped_func = func
		return UndoWrapper
	
		#_____________________________________________________________________
		
	return UndoFunction

	#_____________________________________________________________________

#=========================================================================
"""
These import statements *must* be placed below the UndoCommand function because
decorators are called at import-time to decorate other functions. Project, Instrument
and Event classes all use the UndoCommand decorator. Therefore importing any of those
modules before UndoCommand is defined will cause a cyclic dependency in which
Event depends on UndoSystem and UndoSystem depends on Event. A cyclic import
dependency will stop the program before it even starts.
"""
import ProjectManager, Globals, Utils
import Project, Event, Instrument
import IncrementalSave
import xml.dom.minidom as xml

#=========================================================================

class CancelUndoCommand(Exception):
	"""
	This exception can be thrown by a decorated undo
	function in order to tell the undo system to not
	log the current action. This is useful if something
	in the function fails and the action that would have
	been logged to the undo stack was never actually completed.
	"""
	def __init__(self, result=None):
		"""
		Creates a new instance of CancelUndoCommand.
		
		Parameters:
			result -- value the wrapped function intended to return,
						but failed and called this exception.
		"""
		Exception.__init__(self)
		self.result = result
	
	#_____________________________________________________________________

#=========================================================================

class AtomicUndoAction:
	"""
	Contains several undo commands to be treated as a single undoable operation.
	
	Example:
		When deleting several Instruments at once, an AtomicUndoAction
		containing the commands to resurrect the Instruments will be created.
		When the user requests an undo operation, all of the commands stored
		in this object will be rolled back, making the operation appear to be
		atomic from the user's perspective.
	"""
	
	#_____________________________________________________________________
	
	def __init__(self):
		"""
		Creates a new AtomicUndoAction instance.
		"""
		self.commandList = []
	
	#_____________________________________________________________________
	
	def AddUndoCommand(self, objectString, function, paramList):
		"""
		Adds a new undo command to this AtomicUndoAction.
		
		Example:
			The parameters passed to this function:
				"E2", "Move", [1, 2]
			means
				'Call Move(1, 2)' on the Event with ID=2
		
		Parameters:
			objectString -- the string representing the object and its ID
							(ie "E2" for Event with ID == 2).
			function -- the name of the function to be called on the object.
			paramList -- a list of values to be passed to the function as parameters.
						Key, value parameters are not supported.
		"""
		newTuple = (objectString, function, paramList)
		self.commandList.append(newTuple)
		Globals.debug("LOG COMMAND: ", newTuple, "from", id(self))
	
	#_____________________________________________________________________
	
	def GetUndoCommands(self):
		"""
		Obtains the list of undo commands held by this AtomicUndoAction.
		
		Returns:
			a list of tuples, each of which contains a single undo command.
		"""
		return self.commandList
	
	#_____________________________________________________________________
	
	def StoreToXML(self, doc, node):
		"""
		Stores this instance of AtomicUndoAction into an XML node.
		
		Example:
				doc = xml.Document()
				node = doc.createElement("Action")
				doc.appendChild(node)
				StoreToXml(doc, node)
				
			will save this AtomicUndoAction in doc, inside node.
		
		Parameters:
			doc -- XML document to save this AtomicUndoAction into.
			node -- XML node to store this AtomicUndoAction under.
					This node's name should be "Action".
		"""
		for cmd in self.GetUndoCommands():
			commandXML = doc.createElement("Command")
			node.appendChild(commandXML)
			commandXML.setAttribute("object", cmd[0])
			commandXML.setAttribute("function", cmd[1])
			Utils.StoreListToXML(doc, commandXML, cmd[2], "Parameter")
		
	#_____________________________________________________________________
	
#=========================================================================



#
#	THIS FILE IS PART OF THE JOKOSHER PROJECT AND LICENSED UNDER THE GPL. SEE
#	THE 'COPYING' FILE FOR DETAILS
#
#	TransportManager.py
#	
#	This class handles the current cursor position and the gstreamer
#	bits for rewinding, fast forwarding and seeking.
#	
#
#-------------------------------------------------------------------------------

import pygst
pygst.require("0.10")
import gst
import gobject

#=========================================================================

class TransportManager(gobject.GObject):
	"""
	This class handles the current cursor position and the gstreamer
	bits for rewinding, fast forwarding and seeking.
	"""

	""" Position update rate in Frames Per Second. """
	FPS = 30.
	
	""" Timing resolution - number of ticks ber beat. """
	TICKS_PER_BEAT = 256
	
	""" How many times normal speed the position moves when seeking. """
	SEEK_RATE = 5.
	
	""" Display mode in hours, minutes and seconds. """
	MODE_HOURS_MINS_SECS = 1
	
	""" Display mode in bars, beats and ticks. """
	MODE_BARS_BEATS = 2
	
	"""
	Signals:
		"position" -- The playhead position of the project has changed.
				An optional string is also send which details why the position was changed.
		"transport-mode" -- The mode of measurement for the transport time has changed.
	"""
	
	__gsignals__ = {
		"position"		: ( gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, (gobject.TYPE_STRING,) ),
		"transport-mode"	: ( gobject.SIGNAL_RUN_LAST, gobject.TYPE_NONE, (gobject.TYPE_INT,) )
	}

	#_____________________________________________________________________

	def __init__(self, initialMode, project):
		"""
		Creates a new instance of TransportManager.
		
		Parameters:
			initialMode --the initial mode for the timeline display.
						Possible values:
						MODE_HOURS_MINS_SECS
						MODE_BARS_BEATS
			project -- reference to the current Project.
		"""
		gobject.GObject.__init__(self)
		
		self.project = project
		self.pipeline = self.project.mainpipeline
		self.position = 0
		self.PrevPosition = 0	
		self.isPlaying = False
		self.isPaused = False
		self.isReversing = False
		self.isForwarding = False
		self.UpdateTimeout = False
		self.stopPosition = 0
		self.mode = initialMode

	#_____________________________________________________________________
	
	def Play(self, newAudioState):
		"""
		Called when play button has been pressed (or whilst exporting
		in which case, newAudioState will be set to AUDIO_EXPORTING).
		
		Parameters:
			newAudioState -- new audio state to set the Project to.
		"""
		#the state must be set to paused before playing
		if self.pipeline.get_state(0)[1] != gst.STATE_PAUSED:
			return
			
		if self.isPaused:
			self.isPaused = False
		self.isPlaying = True
		self.project.SetAudioState(newAudioState)
		
		if self.position > 0:
			self.SeekTo(self.position, self.stopPosition)
			#clear stopPosition in case it has been set by previous SeekTo()
			self.stopPosition = 0
		
		self.pipeline.set_state(gst.STATE_PLAYING)
		#for normal playback then we need to start the timeout that will 
		#control the movement of the playhead
		if not self.project.GetIsExporting():
			self.StartUpdateTimeout()
		
	#_____________________________________________________________________
		
	def Pause(self):
		"""
		Pause the playback.
		"""
		self.isPlaying = False
		self.isPaused = True
		self.project.SetAudioState(self.project.AUDIO_PAUSED)
		self.pipeline.set_state(gst.STATE_PAUSED)
	
	#_____________________________________________________________________
		
	def Stop(self):
		"""
		Stops the playback.
		"""
		self.isPlaying = False
		self.isPaused = False
		self.project.SetAudioState(self.project.AUDIO_STOPPED)
		self.SetPosition(0.0, True)
		self.pipeline.set_state(gst.STATE_READY)
		
	#_____________________________________________________________________
		
	def Reverse(self, turnOn):
		"""
		Called when rewind button is
			a) pressed - turnOn = True
			b) released - turnOn = False
			
		Parameters:
			turnOn -- state of the rewind button.
		"""
		if self.isReversing == turnOn:
			#there is no change in reversing state
			return
		
		self.isReversing = turnOn
		if turnOn:
			if self.isPlaying:
				#Pause playback while seeking
				self.pipeline.set_state(gst.STATE_PAUSED)
			self.StartUpdateTimeout()
		else:
			self.SeekTo(self.GetPosition())
			if self.isPlaying:
				#resume playback if it was playing before
				self.pipeline.set_state(gst.STATE_PLAYING)
		
	#_____________________________________________________________________
		
	def Forward(self, turnOn):
		"""
		Called when fast forward button is
			a) pressed - turnOn = True
			b) released - turnOn = False
			
		Parameters:
			turnOn -- state of the fast forward button.
		"""
		if self.isForwarding == turnOn:
			#there is no change in the forwarding state
			return
	
		self.isForwarding = turnOn
		if turnOn:
			if self.isPlaying:
				#Pause playback while seeking
				self.pipeline.set_state(gst.STATE_PAUSED)
			self.StartUpdateTimeout()
		else:
			self.SeekTo(self.GetPosition())
			if self.isPlaying:
				#resume playback if it was playing before
				self.pipeline.set_state(gst.STATE_PLAYING)
		
	#_____________________________________________________________________
	
	def GetPosition(self):
		"""
		Obtain the current playhead position.
		
		Returns:
			the current playhead cursor position.
		"""
		return self.position
	
	#_____________________________________________________________________
	
	def GetPreviousPosition(self):
		"""
		Returns the previous playhead cursor position.
		ie. The value of GetPosition() before the last position update.
		"""
		return self.PrevPosition
	
	#_____________________________________________________________________
	
	def GetPixelPosition(self, offset=None):
		return self._ConvertPositionToPixels(self.position, offset)
	
	#_____________________________________________________________________
	
	def GetPreviousPixelPosition(self, offset=None):
		return self._ConvertPositionToPixels(self.PrevPosition, offset)
	
	#_____________________________________________________________________
	
	def _ConvertPositionToPixels(self, position, offset=None):
		"""
		Parameters:
			position - The play position in seconds.
			offset - The start time in seconds to measure from.
			
		Return:
			the playhead position, measured in pixels from the beginning
			of the visible area (where the view is scrolled to).
		"""
		if offset is None:
			offset = self.project.viewStart
		
		rel_pos = position - offset
		abs_pos = rel_pos * self.project.viewScale
		
		return int(round(abs_pos))
	
	#_____________________________________________________________________
	
	def SetPosition(self, pos, stopAction=False):
		"""
		Change the current position variable.
		
		Considerations:
			Calls emit to trigger response on all classes
			that are connected to this object.
		
		Parameters:
			pos -- new playhead cursor position.
			stopAction -- true if this position change was a result of stopping
					and sending the position back to zero.
		"""
		if self.position != pos:
			self.PrevPosition = self.position
			self.position = pos
			if stopAction:
				self.emit("position", "stop-action")
			else:
				self.emit("position", "")

	#_____________________________________________________________________

	def SetMode(self, mode):
		"""
		In most cases, for undo compatibility, use Project.SetTransportMode().
		
		Parameters:
			mode -- new transport mode to be set.
		"""
		if self.mode != mode:
			self.mode = mode
			self.emit("transport-mode", mode)
		
	#_____________________________________________________________________
	
	def GetPositionAsBarsAndBeats(self):
		"""
		Obtain the current position in bars, beats and ticks.
		
		Returns:
			tuple of the current position as (bar, beats, ticks).
		"""
		mins = self.position / 60.
		if self.project.meter_denom == 8 and (self.project.meter_nom % 3) == 0 and self.project.meter_nom != 3:
			# Compound time
			beats_per_bar = self.project.meter_nom / 3
			beats = int(mins * self.project.bpm / 3)
		else:
			# Simple meter
			beats_per_bar = self.project.meter_nom
			beats = int(mins * self.project.bpm)
		ticks = ((mins - (beats / float(self.project.bpm))) * self.project.bpm) * self.TICKS_PER_BEAT
		bars = int(beats / beats_per_bar)
		beats -= bars * beats_per_bar
		return (bars+1, beats+1, ticks)
		
	#_____________________________________________________________________
	
	def GetPositionAsHoursMinutesSeconds(self):
		"""
		Obtain the current position in hours, minutes and seconds.
		
		Returns:
			tuple of the current position as (hours, minutes, seconds, milliseconds).
		"""
		hours = int(self.position / 3600)
		mins = int((self.position % 3600) / 60)
		secs = int(self.position % 60)
		millis = int((self.position * 1000) % 1000)
		
		return (hours, mins, secs, millis)
	
	#_____________________________________________________________________

	def StartUpdateTimeout(self):
		"""
		Starts the timeout that will control the playhead display.
		"""
		if not self.UpdateTimeout:
			gobject.timeout_add(int(1000/self.FPS), self.OnUpdate)
			self.UpdateTimeout = True
	
	#_____________________________________________________________________
	
	def OnUpdate(self):
		"""
		The timeout callback - called every 1/FPS to move the 
		playhead display on.
		
		Returns:
			True -- pipeline is playing, keep calling this method.
			False -- pipeline is paused or stopped, stop calling this method.
		"""
		if self.isReversing:
			newpos = self.position - self.SEEK_RATE/self.FPS
			self.SetPosition(max(newpos, 0))
		elif self.isForwarding:
			self.SetPosition(self.position + self.SEEK_RATE/self.FPS)
		elif self.isPlaying:
			try:
				#if pipeline should be playing and has not quite 
				#yet started then ignore this time through
				if self.pipeline.get_state(0)[1] == gst.STATE_PAUSED:
					return True
				self.QueryPosition()
			except gst.QueryError:
				pass
		else:
			self.UpdateTimeout = False
			#Prevent the timeout from calling us again
			return False
			
		#Make sure the timeout calls us again
		return True
		
	#_____________________________________________________________________
	
	def SeekTo(self, pos, stopPos=0):
		"""
		Performs a pipeline seek to alter position of the playhead cursor.
		
		Parameters:
			pos -- position to place the playhead cursor.
			stopPos -- new stop position
		"""
		#make sure we cant seek to before the beginning
		pos = max(0, pos)
		if self.isPlaying or self.isPaused:
			#if stopPos is set then pass it to gstreamer here anc clear
			if stopPos:
				self.pipeline.seek( 1.0, gst.FORMAT_TIME, gst.SEEK_FLAG_FLUSH,
						gst.SEEK_TYPE_SET, long(pos * gst.SECOND), 
						gst.SEEK_TYPE_SET, long(stopPos * gst.SECOND))
			else:
				self.pipeline.seek( 1.0, gst.FORMAT_TIME, gst.SEEK_FLAG_FLUSH,
						gst.SEEK_TYPE_SET, long(pos * gst.SECOND), 
						gst.SEEK_TYPE_NONE, 0)
		else:
			#if we haven't done the seek yet (because seek doesn't work while
			#stopped) then we need to save the stop position until we play again
			self.stopPosition = stopPos
		self.SetPosition(pos)
		
	#_____________________________________________________________________
	
	def QueryPosition(self):
		"""
		Reads the current playhead cursor position by querying pipeline.
		"""
		pos = self.pipeline.query_position(gst.FORMAT_TIME)[0]
		self.SetPosition(float(pos) / gst.SECOND)
		
	#_____________________________________________________________________
	
#=========================================================================
